#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for thickener unit example.
"""

from watertap.unit_models.tests.unit_test_harness import UnitTestHarness

import pytest
from pyomo.environ import (
    ConcreteModel,
    units,
)

from idaes.core import FlowsheetBlock

from idaes.models.unit_models.separator import SplittingType

from watertap.core.solvers import get_solver

import idaes.core.util.scaling as iscale


from idaes.core.util.exceptions import (
    ConfigurationError,
)
from watertap.unit_models.thickener import Thickener, ActivatedSludgeModelType
from watertap.property_models.unit_specific.activated_sludge.asm1_properties import (
    ASM1ParameterBlock,
)

from watertap.property_models.unit_specific.activated_sludge.asm2d_properties import (
    ASM2dParameterBlock,
)
from watertap.property_models.unit_specific.activated_sludge.modified_asm2d_properties import (
    ModifiedASM2dParameterBlock,
)
from watertap.costing import WaterTAPCosting
from idaes.core import UnitModelCostingBlock

__author__ = "Alejandro Garciadiego, Adam Atia"

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


# -----------------------------------------------------------------------------
def build_ASM1():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.props = ASM1ParameterBlock()

    m.fs.unit = Thickener(
        property_package=m.fs.props,
        outlet_list=["underflow", "overflow"],
        split_basis=SplittingType.componentFlow,
    )

    m.fs.unit.inlet.flow_vol.fix(300 * units.m**3 / units.day)

    m.fs.unit.inlet.temperature.fix(308.15 * units.K)
    m.fs.unit.inlet.pressure.fix(1 * units.atm)

    m.fs.unit.inlet.conc_mass_comp[0, "S_I"].fix(28.0643 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_S"].fix(0.67336 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_I"].fix(3036.2175 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_S"].fix(63.2392 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_BH"].fix(4442.8377 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_BA"].fix(332.5958 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_P"].fix(1922.8108 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_O"].fix(1.3748 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_NO"].fix(9.1948 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_NH"].fix(0.15845 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_ND"].fix(0.55943 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_ND"].fix(4.7411 * units.mg / units.liter)
    m.fs.unit.inlet.alkalinity.fix(4.5646 * units.mol / units.m**3)

    m.fs.unit.hydraulic_retention_time.fix()
    m.fs.unit.diameter.fix()

    # Set scaling factors for badly scaled variables
    iscale.set_scaling_factor(m.fs.unit.underflow_state[0.0].flow_vol, 1e4)
    iscale.set_scaling_factor(m.fs.unit.underflow_state[0.0].pressure, 1e-6)
    iscale.set_scaling_factor(m.fs.unit.underflow_state[0.0].conc_mass_comp["S_S"], 1e4)
    iscale.set_scaling_factor(
        m.fs.unit.underflow_state[0.0].conc_mass_comp["S_NH"], 1e4
    )
    iscale.set_scaling_factor(
        m.fs.unit.underflow_state[0.0].conc_mass_comp["S_ND"], 1e4
    )
    iscale.set_scaling_factor(m.fs.unit.overflow_state[0.0].pressure, 1e-6)
    iscale.set_scaling_factor(m.fs.unit.overflow_state[0.0].conc_mass_comp["S_S"], 1e4)
    iscale.set_scaling_factor(m.fs.unit.overflow_state[0.0].conc_mass_comp["S_NH"], 1e4)
    iscale.set_scaling_factor(m.fs.unit.overflow_state[0.0].conc_mass_comp["S_ND"], 1e4)
    iscale.set_scaling_factor(m.fs.unit.overflow_state[0.0].conc_mass_comp["X_ND"], 1e4)

    return m


def build_ASM2d():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.props = ASM2dParameterBlock()

    m.fs.unit = Thickener(
        property_package=m.fs.props,
        activated_sludge_model=ActivatedSludgeModelType.ASM2D,
    )

    # NOTE: Concentrations of exactly 0 result in singularities, use EPS instead
    EPS = 1e-8

    m.fs.unit.inlet.flow_vol.fix(300 * units.m**3 / units.day)
    m.fs.unit.inlet.temperature.fix(308.15 * units.K)
    m.fs.unit.inlet.pressure.fix(1 * units.atm)
    m.fs.unit.inlet.conc_mass_comp[0, "S_O2"].fix(7.9707 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_N2"].fix(29.0603 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_NH4"].fix(8.0209 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_NO3"].fix(6.6395 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_PO4"].fix(7.8953 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_F"].fix(0.4748 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_A"].fix(0.0336 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_I"].fix(30 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_I"].fix(1695.7695 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_S"].fix(68.2975 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_H"].fix(1855.5067 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_PAO"].fix(214.5319 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_PP"].fix(63.5316 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_PHA"].fix(2.7381 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_AUT"].fix(118.3582 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_MeOH"].fix(EPS * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_MeP"].fix(EPS * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_TSS"].fix(3525.429 * units.mg / units.liter)
    m.fs.unit.inlet.alkalinity[0].fix(4.6663 * units.mmol / units.liter)

    m.fs.unit.hydraulic_retention_time.fix()
    m.fs.unit.diameter.fix()

    # Set scaling factors for badly scaled variables
    iscale.set_scaling_factor(m.fs.unit.underflow_state[0.0].flow_vol, 1e5)
    iscale.set_scaling_factor(m.fs.unit.underflow_state[0.0].pressure, 1e-6)
    iscale.set_scaling_factor(m.fs.unit.underflow_state[0.0].conc_mass_comp["S_A"], 1e5)
    iscale.set_scaling_factor(m.fs.unit.underflow_state[0.0].conc_mass_comp["S_F"], 1e4)
    iscale.set_scaling_factor(
        m.fs.unit.underflow_state[0.0].conc_mass_comp["X_MeOH"], 1e7
    )
    iscale.set_scaling_factor(
        m.fs.unit.underflow_state[0.0].conc_mass_comp["X_MeP"], 1e7
    )
    iscale.set_scaling_factor(m.fs.unit.overflow_state[0.0].pressure, 1e-6)
    iscale.set_scaling_factor(m.fs.unit.overflow_state[0.0].conc_mass_comp["S_A"], 1e5)
    iscale.set_scaling_factor(m.fs.unit.overflow_state[0.0].conc_mass_comp["S_F"], 1e4)
    iscale.set_scaling_factor(
        m.fs.unit.overflow_state[0.0].conc_mass_comp["X_MeOH"], 1e10
    )
    iscale.set_scaling_factor(
        m.fs.unit.overflow_state[0.0].conc_mass_comp["X_MeP"], 1e10
    )
    iscale.set_scaling_factor(
        m.fs.unit.overflow_state[0.0].conc_mass_comp["X_PHA"], 1e5
    )

    iscale.calculate_scaling_factors(m.fs.unit)

    return m


def build_ASM2d_modified():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.props = ModifiedASM2dParameterBlock()

    m.fs.unit = Thickener(
        property_package=m.fs.props,
        activated_sludge_model=ActivatedSludgeModelType.modified_ASM2D,
    )

    # NOTE: Concentrations of exactly 0 result in singularities, use EPS instead
    EPS = 1e-8

    m.fs.unit.inlet.flow_vol.fix(300 * units.m**3 / units.day)
    m.fs.unit.inlet.temperature.fix(308.15 * units.K)
    m.fs.unit.inlet.pressure.fix(1 * units.atm)
    m.fs.unit.inlet.conc_mass_comp[0, "S_O2"].fix(7.9707 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_N2"].fix(29.0603 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_NH4"].fix(8.0209 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_NO3"].fix(6.6395 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_PO4"].fix(7.8953 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_F"].fix(0.4748 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_A"].fix(0.0336 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_I"].fix(30 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_K"].fix(7 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_Mg"].fix(6 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_IC"].fix(10 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_I"].fix(1695.7695 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_S"].fix(68.2975 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_H"].fix(1855.5067 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_PAO"].fix(214.5319 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_PP"].fix(63.5316 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_PHA"].fix(2.7381 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_AUT"].fix(118.3582 * units.mg / units.liter)

    iscale.calculate_scaling_factors(m.fs.unit)

    m.fs.unit.hydraulic_retention_time.fix()
    m.fs.unit.diameter.fix()

    # Set scaling factors for badly scaled variables
    iscale.set_scaling_factor(m.fs.unit.underflow_state[0.0].flow_vol, 1e5)
    iscale.set_scaling_factor(m.fs.unit.underflow_state[0.0].pressure, 1e-6)
    iscale.set_scaling_factor(m.fs.unit.underflow_state[0.0].conc_mass_comp["S_A"], 1e5)
    iscale.set_scaling_factor(m.fs.unit.underflow_state[0.0].conc_mass_comp["S_F"], 1e4)
    iscale.set_scaling_factor(m.fs.unit.overflow_state[0.0].pressure, 1e-6)
    iscale.set_scaling_factor(m.fs.unit.overflow_state[0.0].conc_mass_comp["S_A"], 1e5)
    iscale.set_scaling_factor(m.fs.unit.overflow_state[0.0].conc_mass_comp["S_F"], 1e4)
    iscale.set_scaling_factor(
        m.fs.unit.overflow_state[0.0].conc_mass_comp["X_PHA"], 1e5
    )

    return m


class TestThickener_ASM1(UnitTestHarness):
    def configure(self):
        m = build_ASM1()

        self.unit_solutions[m.fs.unit.overflow.pressure[0]] = 101325.0
        self.unit_solutions[m.fs.unit.overflow.temperature[0]] = 308.15
        self.unit_solutions[m.fs.unit.overflow.flow_vol[0]] = 0.003115
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "S_I"]] = 0.02806
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "S_S"]] = 0.000673
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "X_I"]] = 0.06768
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "X_S"]] = 0.001409
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "X_P"]] = 0.04286
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "X_BH"]] = 0.099046
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "X_BA"]] = 0.007414
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "S_O"]] = 0.001374
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "S_NO"]] = 0.009194
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "S_NH"]] = 0.0001584
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "S_ND"]] = 0.0005594
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "X_ND"]] = 0.0001056
        self.unit_solutions[m.fs.unit.overflow.alkalinity[0]] = 0.004564

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.inlet.flow_vol[0],
                "out": (
                    m.fs.unit.overflow.flow_vol[0] * m.fs.props.dens_mass
                    + m.fs.unit.underflow.flow_vol[0] * m.fs.props.dens_mass
                )
                / m.fs.props.dens_mass,
            },
            "Check 2": {
                "in": m.fs.unit.inlet.flow_vol[0]
                * sum(
                    m.fs.unit.inlet.conc_mass_comp[0, j] for j in m.fs.props.solute_set
                ),
                "out": sum(
                    m.fs.unit.overflow.flow_vol[0]
                    * m.fs.unit.overflow.conc_mass_comp[0, j]
                    + m.fs.unit.underflow.flow_vol[0]
                    * m.fs.unit.underflow.conc_mass_comp[0, j]
                    for j in m.fs.props.solute_set
                ),
            },
        }

        return m


class TestThickener_ASM2d(UnitTestHarness):
    def configure(self):
        m = build_ASM2d()

        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "S_A"]] = 3.360667e-05
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "S_F"]] = 0.0004748
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "S_I"]] = 0.03
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "S_N2"]] = 0.029060
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "S_NH4"]] = 0.008020
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "S_NO3"]] = 0.006639
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "S_O2"]] = 0.007970
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "S_PO4"]] = 0.007895
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "X_AUT"]] = 0.002490
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "X_H"]] = 0.039036
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "X_I"]] = 0.035676
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "X_MeOH"]] = 0.0
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "X_MeP"]] = 0.0
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "X_PAO"]] = 0.004513
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "X_PHA"]] = (
            5.761182e-05
        )
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "X_PP"]] = 0.001336
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "X_S"]] = 0.001436
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "X_TSS"]] = 0.074169
        self.unit_solutions[m.fs.unit.overflow.alkalinity[0]] = 0.004666

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.inlet.flow_vol[0],
                "out": (
                    m.fs.unit.overflow.flow_vol[0] * m.fs.props.dens_mass
                    + m.fs.unit.underflow.flow_vol[0] * m.fs.props.dens_mass
                )
                / m.fs.props.dens_mass,
            },
            "Check 2": {
                "in": m.fs.unit.inlet.flow_vol[0]
                * sum(
                    m.fs.unit.inlet.conc_mass_comp[0, j] for j in m.fs.props.solute_set
                ),
                "out": sum(
                    m.fs.unit.overflow.flow_vol[0]
                    * m.fs.unit.overflow.conc_mass_comp[0, j]
                    + m.fs.unit.underflow.flow_vol[0]
                    * m.fs.unit.underflow.conc_mass_comp[0, j]
                    for j in m.fs.props.solute_set
                ),
            },
        }

        return m


class TestThickener_ASM2d_modified(UnitTestHarness):
    def configure(self):
        m = build_ASM2d_modified()

        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "S_A"]] = 3.360667e-05
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "S_F"]] = 0.0004748
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "S_I"]] = 0.03
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "S_N2"]] = 0.029060
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "S_NH4"]] = 0.008020
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "S_NO3"]] = 0.006639
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "S_O2"]] = 0.007970
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "S_PO4"]] = 0.007895
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "X_AUT"]] = 0.002490
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "X_H"]] = 0.038882
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "X_I"]] = 0.035535
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "X_PAO"]] = 0.0044955
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "X_PHA"]] = (
            5.761182e-05
        )
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "X_PP"]] = 0.001336
        self.unit_solutions[m.fs.unit.overflow.conc_mass_comp[0, "X_S"]] = 0.001436

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.inlet.flow_vol[0],
                "out": (
                    m.fs.unit.overflow.flow_vol[0] * m.fs.props.dens_mass
                    + m.fs.unit.underflow.flow_vol[0] * m.fs.props.dens_mass
                )
                / m.fs.props.dens_mass,
            },
            "Check 2": {
                "in": m.fs.unit.inlet.flow_vol[0]
                * sum(
                    m.fs.unit.inlet.conc_mass_comp[0, j] for j in m.fs.props.solute_set
                ),
                "out": sum(
                    m.fs.unit.overflow.flow_vol[0]
                    * m.fs.unit.overflow.conc_mass_comp[0, j]
                    + m.fs.unit.underflow.flow_vol[0]
                    * m.fs.unit.underflow.conc_mass_comp[0, j]
                    for j in m.fs.props.solute_set
                ),
            },
        }

        return m


class TestCosting(UnitTestHarness):
    def configure(self):
        m = build_ASM1()

        # Add unit model costing
        m.fs.costing = WaterTAPCosting()

        m.fs.unit.costing = UnitModelCostingBlock(
            flowsheet_costing_block=m.fs.costing,
        )

        iscale.set_scaling_factor(m.fs.unit.costing.capital_cost, 1e-6)

        m.fs.costing.cost_process()

        self.unit_solutions[m.fs.unit.costing.capital_cost] = 220675.79 * 2
        self.unit_solutions[m.fs.unit.electricity_consumption[0]] = 12.5 * 0.01255

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.inlet.flow_vol[0],
                "out": m.fs.unit.underflow.flow_vol[0] + m.fs.unit.overflow.flow_vol[0],
            },
        }

        return m


@pytest.mark.unit
def test_list_error():
    m = ConcreteModel()

    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.props = ASM1ParameterBlock()

    with pytest.raises(
        ConfigurationError,
        match="fs.unit encountered unrecognised "
        "outlet_list. This should not "
        "occur - please use overflow "
        "and underflow as outlets.",
    ):
        m.fs.unit = Thickener(
            property_package=m.fs.props, outlet_list=["outlet1", "outlet2"]
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
GUI configuration for the GAC model.
"""

from pyomo.environ import units as pyunits
from idaes.core import MaterialFlowBasis
from watertap.core.solvers import get_solver
from watertap.property_models.multicomp_aq_sol_prop_pack import DiffusivityCalculation
from watertap.unit_models.gac import (
    FilmTransferCoefficientType,
    SurfaceDiffusionCoefficientType,
)
from watertap.costing.unit_models.gac import ContactorType
from watertap.flowsheets.gac import gac as gac_fs
from idaes_flowsheet_processor.api import FlowsheetInterface, FlowsheetCategory

__author__ = "Hunter Barber"


def export_to_ui():
    """
    Exports the variables, flowsheet build, and solver results to the GUI.
    """
    return FlowsheetInterface(
        name="Granular Activated Carbon (GAC)",
        do_export=export_variables,
        do_build=build_flowsheet,
        do_solve=solve_flowsheet,
        requires_idaes_solver=True,
        category=FlowsheetCategory.wastewater,
        build_options={
            "MaterialFlowBasis": {
                "name": "MaterialFlowBasis",
                "display_name": "Material Flow Basis",
                "values_allowed": [
                    x.name for x in [MaterialFlowBasis.molar, MaterialFlowBasis.mass]
                ],
                "value": MaterialFlowBasis.molar.name,
            },
            "FilmTransferCoefficientType": {
                "name": "FilmTransferCoefficientType",
                "display_name": "Film Transfer Coefficient Type",
                "values_allowed": [x.name for x in FilmTransferCoefficientType],
                "value": FilmTransferCoefficientType.fixed.name,
            },
            "SurfaceDiffusionCoefficientType": {
                "name": "SurfaceDiffusionCoefficientType",
                "display_name": "Surface Diffusion Coefficient Type",
                "values_allowed": [x.name for x in SurfaceDiffusionCoefficientType],
                "value": SurfaceDiffusionCoefficientType.fixed.name,
            },
            "DiffusivityCalculation": {
                "name": "DiffusivityCalculation",
                "display_name": "Diffusivity Calculation",
                "values_allowed": [x.name for x in DiffusivityCalculation],
                "value": DiffusivityCalculation.none.name,
            },
            "ContactorType": {
                "name": "ContactorType",
                "display_name": "Cost Contactor Type",
                "values_allowed": [x.name for x in ContactorType],
                "value": ContactorType.pressure.name,
            },
        },
    )


def export_variables(flowsheet=None, exports=None, build_options=None, **kwargs):
    """
    Exports the variables to the GUI.
    """
    fs = flowsheet
    rounding = 3
    sci_not_rounding = 20
    solute_name = "solute"
    fs.costing.display()
    # input data
    # ---------------------------------------------------------------------
    # solute properties
    category = "Solute properties"
    exports.add(
        obj=fs.properties.mw_comp["solute"],
        name="MW",
        ui_units=pyunits.gram / pyunits.mol,
        display_units="g/mol",
        rounding=rounding,
        description="Molecular weight",
        is_input=True,
        input_category=category,
        is_output=False,
    )
    if (
        build_options["FilmTransferCoefficientType"].value == "calculated"
        or build_options["SurfaceDiffusionCoefficientType"].value == "calculated"
    ):
        if build_options["DiffusivityCalculation"].value == "none":
            exports.add(
                obj=fs.properties.diffus_phase_comp["Liq", "solute"],
                name="Diffusivity",
                ui_units=pyunits.m**2 / pyunits.s,
                display_units="m2/s",
                rounding=sci_not_rounding,
                description="Diffusivity",
                is_input=True,
                input_category=category,
                is_output=False,
            )
        else:
            exports.add(
                obj=fs.properties.molar_volume_phase_comp["Liq", "solute"],
                name="Molar volume",
                ui_units=pyunits.m**3 / pyunits.mol,
                display_units="m3/mol",
                rounding=sci_not_rounding,
                description="Molar volume",
                is_input=True,
                input_category=category,
                is_output=False,
            )
    # ---------------------------------------------------------------------
    # feed conditions
    category = "Feed"
    exports.add(
        obj=fs.feed.properties[0].temperature,
        name="Temperature",
        ui_units=pyunits.kelvin,
        display_units="K",
        rounding=rounding,
        description="Feed temperature",
        is_input=True,
        input_category=category,
        is_output=True,
        output_category=category,
    )
    exports.add(
        obj=fs.feed.properties[0].pressure,
        name="Pressure",
        ui_units=pyunits.bar,
        display_units="bar",
        rounding=rounding,
        description="Feed pressure",
        is_input=True,
        input_category=category,
        is_output=True,
        output_category=category,
    )
    exports.add(
        obj=fs.feed.properties[0].flow_vol_phase["Liq"],
        name="Volumetric flow rate",
        ui_units=pyunits.Mgallon / pyunits.day,
        display_units="MGD",
        rounding=rounding,
        description="Feed volumetric flow rate",
        is_input=False,
        is_output=True,
        output_category=category,
    )
    exports.add(
        obj=fs.feed.properties[0].conc_mass_phase_comp["Liq", solute_name],
        name="Solute concentration",
        ui_units=pyunits.g / pyunits.L,
        display_units="g/L",
        rounding=rounding,
        description="Feed solute concentration",
        is_input=False,
        is_output=True,
        output_category=category,
    )
    if build_options["MaterialFlowBasis"].value == "molar":
        fix_molar = True
        fix_mass = False
    else:
        fix_molar = False
        fix_mass = True
    exports.add(
        obj=fs.feed.properties[0].flow_mol_phase_comp["Liq", "H2O"],
        name="Molar flow rate water",
        ui_units=pyunits.mol / pyunits.s,
        display_units="mol/s",
        rounding=rounding,
        description="Feed molar flow rate of water",
        is_input=fix_molar,
        input_category=category,
        is_output=True,
        output_category=category,
    )
    exports.add(
        obj=fs.feed.properties[0].flow_mol_phase_comp["Liq", solute_name],
        name="Molar flow rate solute",
        ui_units=pyunits.mol / pyunits.s,
        display_units="mol/s",
        rounding=rounding,
        description="Feed molar flow rate of solute",
        is_input=fix_molar,
        input_category=category,
        is_output=True,
        output_category=category,
    )
    exports.add(
        obj=fs.feed.properties[0].flow_mass_phase_comp["Liq", "H2O"],
        name="Mass flow rate water",
        ui_units=pyunits.kg / pyunits.s,
        display_units="kg/s",
        rounding=rounding,
        description="Feed mass flow rate of water",
        is_input=fix_mass,
        input_category=category,
        is_output=True,
        output_category=category,
    )
    exports.add(
        obj=fs.feed.properties[0].flow_mass_phase_comp["Liq", solute_name],
        name="Mass flow rate solute",
        ui_units=pyunits.kg / pyunits.s,
        display_units="kg/s",
        rounding=rounding,
        description="Feed mass flow rate of solute",
        is_input=fix_mass,
        input_category=category,
        is_output=True,
        output_category=category,
    )
    # ---------------------------------------------------------------------
    # product conditions
    category = "Product"
    exports.add(
        obj=fs.product.properties[0].temperature,
        name="Temperature",
        ui_units=pyunits.kelvin,
        display_units="K",
        rounding=rounding,
        description="Product temperature",
        is_input=False,
        is_output=True,
        output_category=category,
    )
    exports.add(
        obj=fs.product.properties[0].pressure,
        name="Pressure",
        ui_units=pyunits.bar,
        display_units="bar",
        rounding=rounding,
        description="Product pressure",
        is_input=False,
        is_output=True,
        output_category=category,
    )
    exports.add(
        obj=fs.product.properties[0].flow_vol_phase["Liq"],
        name="Volumetric flow rate",
        ui_units=pyunits.Mgallon / pyunits.day,
        display_units="MGD",
        rounding=rounding,
        description="Product volumetric flow rate",
        is_input=False,
        is_output=True,
        output_category=category,
    )
    exports.add(
        obj=fs.product.properties[0].conc_mass_phase_comp["Liq", solute_name],
        name="Solute concentration",
        ui_units=pyunits.g / pyunits.L,
        display_units="g/L",
        rounding=rounding,
        description="Product solute concentration",
        is_input=False,
        is_output=True,
        output_category=category,
    )
    exports.add(
        obj=fs.product.properties[0].flow_mol_phase_comp["Liq", "H2O"],
        name="Molar flow rate water",
        ui_units=pyunits.mol / pyunits.s,
        display_units="mol/s",
        rounding=rounding,
        description="Product molar flow rate of water",
        is_input=False,
        is_output=True,
        output_category=category,
    )
    exports.add(
        obj=fs.product.properties[0].flow_mol_phase_comp["Liq", solute_name],
        name="Molar flow rate solute",
        ui_units=pyunits.mol / pyunits.s,
        display_units="mol/s",
        rounding=rounding,
        description="Product molar flow rate of solute",
        is_input=False,
        is_output=True,
        output_category=category,
    )
    # ---------------------------------------------------------------------
    # gac media
    category = "GAC media properties"
    exports.add(
        obj=fs.gac.particle_dens_app,
        name="Apparent density",
        ui_units=pyunits.kg * pyunits.m**-3,
        display_units="kg/m3",
        rounding=rounding,
        description="Apparent density of the GAC media",
        is_input=True,
        input_category=category,
        is_output=False,
    )
    exports.add(
        obj=fs.gac.particle_dens_bulk,
        name="Bulk density",
        ui_units=pyunits.kg * pyunits.m**-3,
        display_units="kg/m3",
        rounding=rounding,
        description="Bulk density of the GAC media in the bed",
        is_input=False,
        is_output=False,
    )
    exports.add(
        obj=fs.gac.particle_dia,
        name="Diameter",
        ui_units=pyunits.mm,
        display_units="mm",
        rounding=rounding,
        description="Particle diameter of the GAC media",
        is_input=True,
        input_category=category,
        is_output=False,
    )
    if build_options["SurfaceDiffusionCoefficientType"].value == "calculated":
        exports.add(
            obj=fs.gac.particle_porosity,
            name="Porosity",
            ui_units=pyunits.dimensionless,
            display_units="-",
            rounding=rounding,
            description="Particle porosity of the GAC media",
            is_input=True,
            input_category=category,
            is_output=False,
        )
    # ---------------------------------------------------------------------
    # adsorption parameters
    category = "Adsorption parameters"
    exports.add(
        obj=fs.gac.freund_k,
        name="Freundlich k",
        ui_units=pyunits.dimensionless,
        display_units="(m3/kg)(1/n)",
        rounding=rounding,
        description="Freundlich isotherm parameter k",
        is_input=True,
        input_category=category,
        is_output=False,
    )
    exports.add(
        obj=fs.gac.freund_ninv,
        name="Freundlich 1/n",
        ui_units=pyunits.dimensionless,
        display_units="-",
        rounding=rounding,
        description="Freundlich isotherm parameter 1/n",
        is_input=True,
        input_category=category,
        is_output=False,
    )
    if build_options["SurfaceDiffusionCoefficientType"].value == "fixed":
        ds_is_input = True
    else:
        ds_is_input = False
        exports.add(
            obj=fs.gac.tort,
            name="Tortuosity",
            ui_units=pyunits.dimensionless,
            display_units="-",
            rounding=rounding,
            description="Tortuosity of the path that the adsorbate must take as compared to the radius",
            is_input=True,
            input_category=category,
            is_output=False,
        )
        exports.add(
            obj=fs.gac.spdfr,
            name="SPDFR",
            ui_units=pyunits.dimensionless,
            display_units="-",
            rounding=rounding,
            description="Surface-to-pore diffusion flux ratio",
            is_input=True,
            input_category=category,
            is_output=False,
        )
    exports.add(
        obj=fs.gac.ds,
        name="Surface diffusion coefficient",
        ui_units=pyunits.m**2 * pyunits.s**-1,
        display_units="m2/s",
        rounding=sci_not_rounding,
        description="Surface diffusion coefficient",
        is_input=ds_is_input,
        input_category=category,
        is_output=False,
    )
    if build_options["FilmTransferCoefficientType"].value == "fixed":
        kf_is_input = True
    else:
        kf_is_input = False
        exports.add(
            obj=fs.gac.shape_correction_factor,
            name="SCF",
            ui_units=pyunits.dimensionless,
            display_units="-",
            rounding=rounding,
            description="Film transfer coefficient shape correction factor",
            is_input=True,
            input_category=category,
            is_output=False,
        )
    exports.add(
        obj=fs.gac.kf,
        name="Film transfer coefficient",
        ui_units=pyunits.m * pyunits.s**-1,
        display_units="m/s",
        rounding=sci_not_rounding,
        description="Liquid phase film transfer coefficient",
        is_input=kf_is_input,
        input_category=category,
        is_output=False,
    )
    # ---------------------------------------------------------------------
    # bed design
    category = "Bed design"
    exports.add(
        obj=fs.gac.velocity_sup,
        name="Superficial velocity",
        ui_units=pyunits.m * pyunits.hour**-1,
        display_units="m/h",
        rounding=rounding,
        description="Superficial velocity",
        is_input=True,
        input_category=category,
        is_output=True,
        output_category=category,
    )
    exports.add(
        obj=fs.gac.velocity_int,
        name="Interstitial velocity",
        ui_units=pyunits.m * pyunits.hour**-1,
        display_units="m/h",
        rounding=rounding,
        description="Interstitial velocity",
        is_input=True,
        input_category=category,
        is_output=True,
        output_category=category,
    )
    exports.add(
        obj=fs.gac.bed_length,
        name="Bed length",
        ui_units=pyunits.m,
        display_units="m",
        rounding=rounding,
        description="Length of the GAC bed",
        is_input=True,
        input_category=category,
        is_output=True,
        output_category=category,
    )
    exports.add(
        obj=fs.gac.bed_diameter,
        name="Bed diameter",
        ui_units=pyunits.m,
        display_units="m",
        rounding=rounding,
        description="Effective bed diameter as if only once bed was present",
        is_input=True,
        input_category=category,
        is_output=True,
        output_category=category,
        is_readonly=True,
    )
    exports.add(
        obj=fs.gac.bed_area,
        name="Bed area",
        ui_units=pyunits.m**2,
        display_units="m2",
        rounding=rounding,
        description="Effective total bed area for flow",
        is_input=True,
        input_category=category,
        is_output=True,
        output_category=category,
    )
    exports.add(
        obj=fs.gac.bed_volume,
        name="Bed volume",
        ui_units=pyunits.m**3,
        display_units="m3",
        rounding=rounding,
        description="Effective total bed volume",
        is_input=True,
        input_category=category,
        is_output=True,
        output_category=category,
    )
    exports.add(
        obj=fs.gac.ebct,
        name="EBCT",
        ui_units=pyunits.min,
        display_units="min",
        rounding=rounding,
        description="Empty bed contact time",
        is_input=True,
        input_category=category,
        is_output=True,
        output_category=category,
    )
    exports.add(
        obj=fs.gac.residence_time,
        name="Residence time",
        ui_units=pyunits.min,
        display_units="min",
        rounding=rounding,
        description="Residence time",
        is_input=True,
        input_category=category,
        is_output=True,
        output_category=category,
    )
    exports.add(
        obj=fs.gac.bed_voidage,
        name="Bed voidage",
        ui_units=pyunits.dimensionless,
        display_units="-",
        rounding=rounding,
        description="Voidage of the GAC bed",
        is_input=True,
        input_category=category,
        is_output=True,
        output_category=category,
    )
    exports.add(
        obj=fs.gac.bed_mass_gac,
        name="Media mass",
        ui_units=pyunits.kg,
        display_units="kg",
        rounding=rounding,
        description="Mass of media in the bed",
        is_input=True,
        input_category=category,
        is_output=True,
        output_category=category,
    )
    # ---------------------------------------------------------------------
    # system performance
    category = "System performance"
    exports.add(
        obj=fs.gac.conc_ratio_replace,
        name="Breakthrough concentration ratio",
        ui_units=pyunits.dimensionless,
        display_units="-",
        rounding=rounding,
        description="Concentration ratio at breakthrough, corresponding to the time of replacing the media",
        is_input=True,
        input_category=category,
        is_output=True,
        output_category=category,
    )
    exports.add(
        obj=fs.gac.conc_ratio_avg,
        name="Average concentration ratio",
        ui_units=pyunits.dimensionless,
        display_units="-",
        rounding=rounding,
        description="Average concentration ratio during operational time",
        is_input=True,
        input_category=category,
        is_output=True,
        output_category=category,
    )
    exports.add(
        obj=fs.gac.operational_time,
        name="Breakthrough time",
        ui_units=pyunits.day,
        display_units="days",
        rounding=rounding,
        description="Time operating until breakthrough",
        is_input=True,
        input_category=category,
        is_output=True,
        output_category=category,
    )
    exports.add(
        obj=fs.gac.bed_volumes_treated,
        name="BVT",
        ui_units=pyunits.dimensionless,
        display_units="-",
        rounding=rounding,
        description="Bed volumes treated",
        is_input=True,
        input_category=category,
        is_output=True,
        output_category=category,
    )
    exports.add(
        obj=fs.gac.mass_adsorbed,
        name="Solute adsorption rate",
        ui_units=pyunits.kg,
        display_units="kg",
        rounding=rounding,
        description="Mass adsorption rate of solute during operational time",
        is_input=True,
        input_category=category,
        is_output=True,
        output_category=category,
    )
    exports.add(
        obj=fs.gac.gac_usage_rate,
        name="Media usage rate",
        ui_units=pyunits.kg * pyunits.day**-1,
        display_units="kg/day",
        rounding=rounding,
        description="Mass usage rate of media during operational time",
        is_input=True,
        input_category=category,
        is_output=True,
        output_category=category,
    )
    # ---------------------------------------------------------------------
    # empirical parameters
    category = "Empirical parameters"
    exports.add(
        obj=fs.gac.a0,
        name="Stanton parameter 0",
        ui_units=pyunits.dimensionless,
        display_units="-",
        rounding=6,
        description="Stanton parameter 0",
        is_input=True,
        input_category=category,
        is_output=False,
    )
    exports.add(
        obj=fs.gac.a1,
        name="Stanton parameter 1",
        ui_units=pyunits.dimensionless,
        display_units="-",
        rounding=6,
        description="Stanton parameter 1",
        is_input=True,
        input_category=category,
        is_output=False,
    )
    exports.add(
        obj=fs.gac.b0,
        name="Throughput parameter 0",
        ui_units=pyunits.dimensionless,
        display_units="-",
        rounding=6,
        description="Throughput parameter 0",
        is_input=True,
        input_category=category,
        is_output=False,
    )
    exports.add(
        obj=fs.gac.b1,
        name="Throughput parameter 1",
        ui_units=pyunits.dimensionless,
        display_units="-",
        rounding=6,
        description="Throughput parameter 1",
        is_input=True,
        input_category=category,
        is_output=False,
    )
    exports.add(
        obj=fs.gac.b2,
        name="Throughput parameter 2",
        ui_units=pyunits.dimensionless,
        display_units="-",
        rounding=6,
        description="Throughput parameter 2",
        is_input=True,
        input_category=category,
        is_output=False,
    )
    exports.add(
        obj=fs.gac.b3,
        name="Throughput parameter 3",
        ui_units=pyunits.dimensionless,
        display_units="-",
        rounding=6,
        description="Throughput parameter 3",
        is_input=True,
        input_category=category,
        is_output=False,
    )
    exports.add(
        obj=fs.gac.b4,
        name="Throughput parameter 4",
        ui_units=pyunits.dimensionless,
        display_units="-",
        rounding=6,
        description="Throughput parameter 4",
        is_input=True,
        input_category=category,
        is_output=False,
    )
    # ---------------------------------------------------------------------
    # costing options
    category = "Costing options"
    exports.add(
        obj=fs.costing.electricity_cost,
        name="Electricity cost",
        ui_units=fs.costing.base_currency / pyunits.kWh,
        display_units="$/kW",
        rounding=rounding,
        description="Electricity cost",
        is_input=True,
        input_category=category,
        is_output=True,
        output_category=category,
    )
    if build_options["ContactorType"].value == "pressure":
        unit_model_costing = fs.costing.gac_pressure
    else:
        unit_model_costing = fs.costing.gac_gravity
    exports.add(
        obj=unit_model_costing.regen_frac,
        name="Regeneration fraction",
        ui_units=pyunits.dimensionless,
        display_units="-",
        rounding=rounding,
        description="Fraction of media that can be regenerated after breakthrough (remainder is replaced)",
        is_input=True,
        input_category=category,
        is_output=True,
        output_category=category,
    )
    exports.add(
        obj=unit_model_costing.num_contactors_op,
        name="Number of operating beds",
        ui_units=pyunits.dimensionless,
        display_units="-",
        rounding=rounding,
        description="Number of beds in operation at any steady state time",
        is_input=True,
        input_category=category,
        is_output=True,
        output_category=category,
    )
    exports.add(
        obj=unit_model_costing.num_contactors_redundant,
        name="Number of redundant beds",
        ui_units=pyunits.dimensionless,
        display_units="-",
        rounding=rounding,
        description="Number of beds offline (redundant) at any steady state time",
        is_input=True,
        input_category=category,
        is_output=True,
        output_category=category,
    )
    # ---------------------------------------------------------------------
    # cost results
    category = "Cost Results"
    exports.add(
        obj=fs.costing.aggregate_capital_cost,
        name="Capital costs",
        ui_units=fs.costing.base_currency,
        display_units="$",
        rounding=0,
        description="Capital costs",
        is_input=False,
        is_output=True,
        output_category=category,
    )
    exports.add(
        obj=fs.costing.aggregate_fixed_operating_cost,
        name="Fixed operating costs",
        ui_units=fs.costing.base_currency / pyunits.year,
        display_units="$/yr",
        rounding=0,
        description="Fixed operating costs",
        is_input=False,
        is_output=True,
        output_category=category,
    )
    exports.add(
        obj=fs.costing.aggregate_flow_costs["electricity"],
        name="Electricity costs",
        ui_units=fs.costing.base_currency / pyunits.year,
        display_units="$/yr",
        rounding=0,
        description="Electricity costs",
        is_input=False,
        is_output=True,
        output_category=category,
    )
    exports.add(
        obj=fs.costing.LCOW,
        name="LCOW",
        ui_units=fs.costing.base_currency / pyunits.meter**3,
        display_units="$/yr",
        rounding=rounding,
        description="Levelized cost of water",
        is_input=False,
        is_output=True,
        output_category=category,
    )


def build_flowsheet(build_options=None, **kwargs):
    """
    Build and solve the initial flowsheet.
    """

    if build_options is not None:
        m = gac_fs.build(
            material_flow_basis=build_options["MaterialFlowBasis"].value,
            film_transfer_coefficient_type=build_options[
                "FilmTransferCoefficientType"
            ].value,
            surface_diffusion_coefficient_type=build_options[
                "SurfaceDiffusionCoefficientType"
            ].value,
            diffusivity_calculation=build_options["DiffusivityCalculation"].value,
            cost_contactor_type=build_options["ContactorType"].value,
        )
    else:
        m = gac_fs.build()
    gac_fs.initialize(m)

    return m


def solve_flowsheet(flowsheet=None):
    """
    Solves the flowsheet.
    """

    fs = flowsheet
    solver = get_solver()
    res = gac_fs.optimize(fs, solver)

    return res



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
__author__ = "Alejandro Garciadiego, Adam Atia"

import pyomo.environ as pyo
from pyomo.environ import (
    units,
)
from idaes.core import FlowsheetBlock
import idaes.core.util.scaling as iscale
from watertap.core.solvers import get_solver
import idaes.logger as idaeslog
from watertap.unit_models.anaerobic_digester import AD
from watertap.property_models.unit_specific.anaerobic_digestion.adm1_properties import (
    ADM1ParameterBlock,
)
from watertap.property_models.unit_specific.anaerobic_digestion.adm1_properties_vapor import (
    ADM1_vaporParameterBlock,
)
from watertap.property_models.unit_specific.anaerobic_digestion.adm1_reactions import (
    ADM1ReactionParameterBlock,
)


def build_flowsheet():
    m = pyo.ConcreteModel()

    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.props = ADM1ParameterBlock()
    m.fs.props_vap = ADM1_vaporParameterBlock()
    m.fs.rxn_props = ADM1ReactionParameterBlock(property_package=m.fs.props)

    m.fs.R1 = AD(
        liquid_property_package=m.fs.props,
        vapor_property_package=m.fs.props_vap,
        reaction_package=m.fs.rxn_props,
        has_heat_transfer=True,
        has_pressure_change=False,
    )

    # Feed conditions based on manual mass balance of inlet and recycle streams
    m.fs.R1.inlet.flow_vol.fix(170 * pyo.units.m**3 / pyo.units.day)
    m.fs.R1.inlet.temperature.fix(308.15 * pyo.units.K)
    m.fs.R1.inlet.pressure.fix(1 * pyo.units.atm)
    m.fs.R1.inlet.conc_mass_comp[0, "S_su"].fix(10 * pyo.units.mg / pyo.units.liter)
    m.fs.R1.inlet.conc_mass_comp[0, "S_aa"].fix(1 * pyo.units.mg / pyo.units.liter)
    m.fs.R1.inlet.conc_mass_comp[0, "S_fa"].fix(1 * pyo.units.mg / pyo.units.liter)
    m.fs.R1.inlet.conc_mass_comp[0, "S_va"].fix(1 * pyo.units.mg / pyo.units.liter)
    m.fs.R1.inlet.conc_mass_comp[0, "S_bu"].fix(1 * pyo.units.mg / pyo.units.liter)
    m.fs.R1.inlet.conc_mass_comp[0, "S_pro"].fix(1 * pyo.units.mg / pyo.units.liter)
    m.fs.R1.inlet.conc_mass_comp[0, "S_ac"].fix(1 * pyo.units.mg / pyo.units.liter)
    m.fs.R1.inlet.conc_mass_comp[0, "S_h2"].fix(1e-5 * pyo.units.mg / pyo.units.liter)
    m.fs.R1.inlet.conc_mass_comp[0, "S_ch4"].fix(1e-2 * pyo.units.mg / pyo.units.liter)
    m.fs.R1.inlet.conc_mass_comp[0, "S_IC"].fix(
        40 * units.mmol / units.liter * 12 * units.mg / units.mmol
    )
    m.fs.R1.inlet.conc_mass_comp[0, "S_IN"].fix(
        10 * units.mmol / units.liter * 14 * units.mg / units.mmol
    )
    m.fs.R1.inlet.conc_mass_comp[0, "S_I"].fix(20 * pyo.units.mg / pyo.units.liter)

    m.fs.R1.inlet.conc_mass_comp[0, "X_c"].fix(2000 * pyo.units.mg / pyo.units.liter)
    m.fs.R1.inlet.conc_mass_comp[0, "X_ch"].fix(5000 * pyo.units.mg / pyo.units.liter)
    m.fs.R1.inlet.conc_mass_comp[0, "X_pr"].fix(20000 * pyo.units.mg / pyo.units.liter)
    m.fs.R1.inlet.conc_mass_comp[0, "X_li"].fix(5000 * pyo.units.mg / pyo.units.liter)
    m.fs.R1.inlet.conc_mass_comp[0, "X_su"].fix(0.0 * pyo.units.mg / pyo.units.liter)
    m.fs.R1.inlet.conc_mass_comp[0, "X_aa"].fix(10 * pyo.units.mg / pyo.units.liter)
    m.fs.R1.inlet.conc_mass_comp[0, "X_fa"].fix(10 * pyo.units.mg / pyo.units.liter)
    m.fs.R1.inlet.conc_mass_comp[0, "X_c4"].fix(10 * pyo.units.mg / pyo.units.liter)
    m.fs.R1.inlet.conc_mass_comp[0, "X_pro"].fix(10 * pyo.units.mg / pyo.units.liter)
    m.fs.R1.inlet.conc_mass_comp[0, "X_ac"].fix(10 * pyo.units.mg / pyo.units.liter)
    m.fs.R1.inlet.conc_mass_comp[0, "X_h2"].fix(10 * pyo.units.mg / pyo.units.liter)
    m.fs.R1.inlet.conc_mass_comp[0, "X_I"].fix(25000 * pyo.units.mg / pyo.units.liter)

    m.fs.R1.inlet.cations[0].fix(40 * pyo.units.mmol / pyo.units.liter)
    m.fs.R1.inlet.anions[0].fix(20 * pyo.units.mmol / pyo.units.liter)

    m.fs.R1.volume_liquid.fix(3400 * pyo.units.m**3)
    m.fs.R1.volume_vapor.fix(300 * pyo.units.m**3)

    m.fs.R1.liquid_outlet.temperature.fix(308.15 * pyo.units.K)

    iscale.calculate_scaling_factors(m)

    # TO DO: Fix initialization
    m.fs.R1.initialize(outlvl=idaeslog.INFO_HIGH)

    solver = get_solver()

    results = solver.solve(m, tee=True)

    pyo.assert_optimal_termination(results)

    return m, results



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

from pyomo.environ import (
    ConcreteModel,
)

from idaes.core import FlowsheetBlock
import watertap.property_models.water_prop_pack as props_w
import watertap.property_models.seawater_prop_pack as props_sw
from watertap.unit_models.steam_heater_0D import SteamHeater0D, Mode
from watertap.core.solvers import get_solver
from watertap.unit_models.mvc.components.lmtd_chen_callback import (
    delta_temperature_chen_callback,
)
from idaes.models.unit_models.heat_exchanger import (
    HeatExchangerFlowPattern,
)
import idaes.core.util.scaling as iscale
import pytest
from watertap.unit_models.tests.unit_test_harness import UnitTestHarness


solver = get_solver()


def build(mode, estimate_cooling_water=False):
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.steam_properties = props_w.WaterParameterBlock()
    m.fs.cold_side_properties = props_sw.SeawaterParameterBlock()

    m.fs.unit = SteamHeater0D(
        hot_side_name="hot",
        cold_side_name="cold",
        hot={
            "property_package": m.fs.steam_properties,
        },
        cold={
            "property_package": m.fs.cold_side_properties,
        },
        delta_temperature_callback=delta_temperature_chen_callback,
        flow_pattern=HeatExchangerFlowPattern.countercurrent,
        mode=mode,
        estimate_cooling_water=estimate_cooling_water,
    )

    m.fs.unit.hot_side_inlet.flow_mass_phase_comp[0, "Vap", "H2O"].set_value(1)
    m.fs.unit.hot_side_inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(0)
    m.fs.unit.hot_side_inlet.temperature.fix(273.15 + 140)
    m.fs.unit.hot_side_inlet.pressure[0].fix(201325)
    m.fs.unit.cold_side_inlet.pressure.fix(101325)
    m.fs.unit.cold_side_inlet.temperature.fix(273.15 + 25)
    total_flow_mass = 10
    TDS_mass_frac = 0.035
    m.fs.unit.cold_side_inlet.flow_mass_phase_comp[0, "Liq", "TDS"].fix(
        TDS_mass_frac * total_flow_mass
    )
    m.fs.unit.cold_side_inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
        (1 - TDS_mass_frac) * total_flow_mass
    )
    m.fs.unit.area.fix(10)
    m.fs.unit.overall_heat_transfer_coefficient.fix(2e3)

    iscale.calculate_scaling_factors(m.fs.unit)

    return m


@pytest.mark.requires_idaes_solver
class TestSteamHeater0D(UnitTestHarness):
    def configure(self):
        m = build(Mode.HEATER)

        self.unit_solutions[
            m.fs.unit.hot_side_inlet.flow_mass_phase_comp[0, "Vap", "H2O"]
        ] = 0.6913572208080987
        self.unit_solutions[m.fs.unit.hot_side_outlet.temperature[0]] = (
            381.3881358453978
        )
        self.unit_solutions[m.fs.unit.cold_side_outlet.temperature[0]] = (
            338.631650369243
        )

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.hot_side_inlet.flow_mass_phase_comp[0, "Vap", "H2O"]
                + m.fs.unit.hot_side_inlet.flow_mass_phase_comp[0, "Liq", "H2O"]
                + m.fs.unit.cold_side_inlet.flow_mass_phase_comp[0, "Liq", "H2O"]
                + m.fs.unit.cold_side_inlet.flow_mass_phase_comp[0, "Liq", "TDS"],
                "out": m.fs.unit.hot_side_outlet.flow_mass_phase_comp[0, "Vap", "H2O"]
                + m.fs.unit.hot_side_outlet.flow_mass_phase_comp[0, "Liq", "H2O"]
                + m.fs.unit.cold_side_outlet.flow_mass_phase_comp[0, "Liq", "TDS"]
                + m.fs.unit.cold_side_outlet.flow_mass_phase_comp[0, "Liq", "H2O"],
            },
        }
        return m


@pytest.mark.requires_idaes_solver
class TestCondenserNoEstimation(UnitTestHarness):
    def configure(self):
        m = build(Mode.CONDENSER, estimate_cooling_water=False)
        m.fs.unit.hot_side_inlet.flow_mass_phase_comp[0, "Vap", "H2O"].fix(0.5)
        m.fs.unit.area.unfix()

        self.unit_solutions[m.fs.unit.hot_side_outlet.temperature[0]] = (
            375.3825125826731
        )
        self.unit_solutions[m.fs.unit.cold_side_outlet.temperature[0]] = (
            327.67799780959757
        )
        self.unit_solutions[m.fs.unit.area] = 7.089052938081363
        self.unit_solutions[
            m.fs.unit.hot_side_inlet.flow_mass_phase_comp[0, "Vap", "H2O"]
        ] = 0.5

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.hot_side_inlet.flow_mass_phase_comp[0, "Vap", "H2O"]
                + m.fs.unit.hot_side_inlet.flow_mass_phase_comp[0, "Liq", "H2O"]
                + m.fs.unit.cold_side_inlet.flow_mass_phase_comp[0, "Liq", "H2O"]
                + m.fs.unit.cold_side_inlet.flow_mass_phase_comp[0, "Liq", "TDS"],
                "out": m.fs.unit.hot_side_outlet.flow_mass_phase_comp[0, "Vap", "H2O"]
                + m.fs.unit.hot_side_outlet.flow_mass_phase_comp[0, "Liq", "H2O"]
                + m.fs.unit.cold_side_outlet.flow_mass_phase_comp[0, "Liq", "TDS"]
                + m.fs.unit.cold_side_outlet.flow_mass_phase_comp[0, "Liq", "H2O"],
            },
        }

        return m


@pytest.mark.requires_idaes_solver
class TestCondenserwithEstimation(UnitTestHarness):
    def configure(self):
        m = build(Mode.CONDENSER, estimate_cooling_water=True)
        m.fs.unit.area.unfix()

        outlet_temperature = 340
        m.fs.unit.cold_side_outlet.temperature[0].fix(outlet_temperature)
        m.fs.unit.hot_side_inlet.flow_mass_phase_comp[0, "Vap", "H2O"].fix(0.5)
        total_flow_mass = 12
        TDS_mass_frac = 0.035
        m.fs.unit.cold_side_inlet.flow_mass_phase_comp[0, "Liq", "H2O"].unfix()
        m.fs.unit.cold_side_inlet.flow_mass_phase_comp[0, "Liq", "TDS"].unfix()
        m.fs.unit.cold_side_inlet.flow_mass_phase_comp[0, "Liq", "TDS"].fix(
            TDS_mass_frac * total_flow_mass
        )
        m.fs.unit.cold_side_inlet.flow_mass_phase_comp[0, "Liq", "H2O"].set_value(
            (1 - TDS_mass_frac) * total_flow_mass
        )

        self.unit_solutions[m.fs.unit.hot_side_outlet.temperature[0]] = (
            373.69526509227114
        )
        self.unit_solutions[m.fs.unit.cold_side_outlet.temperature[0]] = 340
        self.unit_solutions[m.fs.unit.area] = 7.774904144204368
        self.unit_solutions[
            m.fs.unit.cold_side_inlet.flow_mass_phase_comp[0, "Liq", "TDS"]
        ] = 0.24841897743647612
        self.unit_solutions[
            m.fs.unit.cold_side_inlet.flow_mass_phase_comp[0, "Liq", "H2O"]
        ] = 6.849266092188223
        self.unit_solutions[
            m.fs.unit.cold_side.properties_in[0].mass_frac_phase_comp["Liq", "TDS"]
        ] = 0.035

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.hot_side_inlet.flow_mass_phase_comp[0, "Vap", "H2O"]
                + m.fs.unit.hot_side_inlet.flow_mass_phase_comp[0, "Liq", "H2O"]
                + m.fs.unit.cold_side_inlet.flow_mass_phase_comp[0, "Liq", "H2O"]
                + m.fs.unit.cold_side_inlet.flow_mass_phase_comp[0, "Liq", "TDS"],
                "out": m.fs.unit.hot_side_outlet.flow_mass_phase_comp[0, "Vap", "H2O"]
                + m.fs.unit.hot_side_outlet.flow_mass_phase_comp[0, "Liq", "H2O"]
                + m.fs.unit.cold_side_outlet.flow_mass_phase_comp[0, "Liq", "TDS"]
                + m.fs.unit.cold_side_outlet.flow_mass_phase_comp[0, "Liq", "H2O"],
            },
        }

        return m



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for zero-order autothermal hydrothermal liquefaction model
"""
import pytest
import os


from pyomo.environ import (
    Block,
    ConcreteModel,
    Constraint,
    value,
    Var,
    assert_optimal_termination,
)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from watertap.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.core import UnitModelCostingBlock

from watertap.unit_models.zero_order import ATHTLZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.costing.zero_order_costing import ZeroOrderCosting

solver = get_solver()


class TestATHTLZO:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(
            solute_list=[
                "organic_solid",
                "organic_liquid",
                "inorganic_solid",
                "carbon_dioxide",
            ]
        )

        m.fs.unit = ATHTLZO(property_package=m.fs.params, database=m.db)

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(400)
        m.fs.unit.inlet.flow_mass_comp[0, "organic_solid"].fix(71.2)
        m.fs.unit.inlet.flow_mass_comp[0, "organic_liquid"].fix(0)
        m.fs.unit.inlet.flow_mass_comp[0, "inorganic_solid"].fix(28.8)
        m.fs.unit.inlet.flow_mass_comp[0, "carbon_dioxide"].fix(0)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database == model.db

        assert isinstance(model.fs.unit.catalyst_dosage, Var)
        assert isinstance(model.fs.unit.flow_mass_in, Var)
        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_mass, Var)
        assert isinstance(model.fs.unit.electricity_consumption, Constraint)
        assert isinstance(model.fs.unit.cons_flow_mass, Constraint)

    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters(
            "autothermal_hydrothermal_liquefaction"
        )

        model.fs.unit.load_parameters_from_database(use_default_removal=True)

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert (
            model.fs.unit.recovery_frac_mass_H2O[0].value
            == data["recovery_frac_mass_H2O"]["value"]
        )

        for (t, j), v in model.fs.unit.removal_frac_mass_comp.items():
            assert v.fixed
            if j not in data["removal_frac_mass_comp"].keys():
                assert v.value == data["default_removal_frac_mass_comp"]["value"]
            else:
                assert v.value == data["removal_frac_mass_comp"][j]["value"]

        assert model.fs.unit.catalyst_dosage.fixed
        assert model.fs.unit.catalyst_dosage.value == data["catalyst_dosage"]["value"]

        assert model.fs.unit.energy_electric_flow_mass.fixed
        assert (
            model.fs.unit.energy_electric_flow_mass.value
            == data["energy_electric_flow_mass"]["value"]
        )

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert pytest.approx(0.5, rel=1e-5) == value(
            model.fs.unit.properties_in[0].flow_vol
        )
        assert pytest.approx(1800, rel=1e-5) == value(model.fs.unit.flow_mass_in[0])
        assert pytest.approx(142.4, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["organic_solid"]
        )
        assert pytest.approx(57.6, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["inorganic_solid"]
        )

        assert pytest.approx(0.4965, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].flow_vol
        )
        assert pytest.approx(14.3403, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["organic_solid"]
        )
        assert pytest.approx(122.0163, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["organic_liquid"]
        )

        assert pytest.approx(0.003499, rel=1e-3) == value(
            model.fs.unit.properties_byproduct[0].flow_vol
        )
        assert pytest.approx(6.800226e-12, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["organic_solid"]
        )
        assert pytest.approx(1000, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["carbon_dioxide"]
        )
        assert pytest.approx(15678, rel=1e-5) == value(model.fs.unit.electricity[0])

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(
                value(
                    model.fs.unit.inlet.flow_mass_comp[0, j]
                    + sum(
                        model.fs.unit.generation_rxn_comp[0, r, j]
                        for r in model.fs.unit.reaction_set
                    )
                    - model.fs.unit.treated.flow_mass_comp[0, j]
                    - model.fs.unit.byproduct.flow_mass_comp[0, j]
                )
            )

    @pytest.mark.component
    def test_report(self, model):

        model.fs.unit.report()


def test_costing():
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = WaterParameterBlock(
        solute_list=[
            "organic_solid",
            "organic_liquid",
            "inorganic_solid",
            "carbon_dioxide",
        ]
    )

    source_file = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "..",
        "..",
        "..",
        "data",
        "techno_economic",
        "supercritical_sludge_to_gas_global_costing.yaml",
    )

    m.fs.costing = ZeroOrderCosting(case_study_definition=source_file)

    m.fs.unit = ATHTLZO(property_package=m.fs.params, database=m.db)

    m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(400)
    m.fs.unit.inlet.flow_mass_comp[0, "organic_solid"].fix(71.2)
    m.fs.unit.inlet.flow_mass_comp[0, "organic_liquid"].fix(0)
    m.fs.unit.inlet.flow_mass_comp[0, "inorganic_solid"].fix(28.8)
    m.fs.unit.inlet.flow_mass_comp[0, "carbon_dioxide"].fix(0)
    m.fs.unit.load_parameters_from_database(use_default_removal=True)
    assert degrees_of_freedom(m.fs.unit) == 0

    m.fs.unit.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    assert isinstance(m.fs.costing.autothermal_hydrothermal_liquefaction, Block)
    assert isinstance(
        m.fs.costing.autothermal_hydrothermal_liquefaction.installation_factor_reactor,
        Var,
    )
    assert isinstance(
        m.fs.costing.autothermal_hydrothermal_liquefaction.equipment_cost_reactor, Var
    )
    assert isinstance(
        m.fs.costing.autothermal_hydrothermal_liquefaction.base_flowrate_reactor, Var
    )
    assert isinstance(
        m.fs.costing.autothermal_hydrothermal_liquefaction.scaling_exponent_reactor, Var
    )
    assert isinstance(
        m.fs.costing.autothermal_hydrothermal_liquefaction.installation_factor_pump, Var
    )
    assert isinstance(
        m.fs.costing.autothermal_hydrothermal_liquefaction.equipment_cost_pump, Var
    )
    assert isinstance(
        m.fs.costing.autothermal_hydrothermal_liquefaction.base_flowrate_pump, Var
    )
    assert isinstance(
        m.fs.costing.autothermal_hydrothermal_liquefaction.scaling_exponent_pump, Var
    )
    assert isinstance(
        m.fs.costing.autothermal_hydrothermal_liquefaction.installation_factor_other,
        Var,
    )
    assert isinstance(
        m.fs.costing.autothermal_hydrothermal_liquefaction.equipment_cost_other, Var
    )
    assert isinstance(
        m.fs.costing.autothermal_hydrothermal_liquefaction.base_flowrate_other, Var
    )
    assert isinstance(
        m.fs.costing.autothermal_hydrothermal_liquefaction.scaling_exponent_other, Var
    )
    assert isinstance(
        m.fs.costing.autothermal_hydrothermal_liquefaction.installation_factor_solid_filter,
        Var,
    )
    assert isinstance(
        m.fs.costing.autothermal_hydrothermal_liquefaction.equipment_cost_solid_filter,
        Var,
    )
    assert isinstance(
        m.fs.costing.autothermal_hydrothermal_liquefaction.base_flowrate_solid_filter,
        Var,
    )
    assert isinstance(
        m.fs.costing.autothermal_hydrothermal_liquefaction.scaling_exponent_solid_filter,
        Var,
    )
    assert isinstance(
        m.fs.costing.autothermal_hydrothermal_liquefaction.installation_factor_heat,
        Var,
    )
    assert isinstance(
        m.fs.costing.autothermal_hydrothermal_liquefaction.equipment_cost_heat, Var
    )
    assert isinstance(
        m.fs.costing.autothermal_hydrothermal_liquefaction.base_flowrate_heat, Var
    )
    assert isinstance(
        m.fs.costing.autothermal_hydrothermal_liquefaction.scaling_exponent_heat, Var
    )

    assert isinstance(m.fs.unit.costing.capital_cost, Var)
    assert isinstance(m.fs.unit.costing.capital_cost_constraint, Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit) == 0
    initialization_tester(m)
    results = solver.solve(m)
    assert_optimal_termination(results)

    assert m.fs.unit.electricity[0] in m.fs.costing._registered_flows["electricity"]
    assert (
        m.fs.unit.catalyst_flow[0] in m.fs.costing._registered_flows["catalyst_ATHTL"]
    )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

from pyomo.environ import (
    Var,
    Constraint,
    units as pyunits,
)

__author__ = "Alexander V. Dudchenko"


def cost_stream(costing_block, stream, stream_cost_value=0, opt_name=None):
    stream.base_cost = Var(
        initialize=stream_cost_value, units=costing_block.base_currency / pyunits.m**3
    )
    stream.base_cost.fix()
    stream.annual_cost = Var(
        initialize=1, units=costing_block.base_currency / pyunits.year
    )
    stream.absolute_cost_eq = Constraint(
        expr=stream.annual_cost
        == stream.base_cost
        * pyunits.convert(
            stream.properties[0].flow_vol_phase["Liq"],
            to_units=pyunits.m**3 / pyunits.year,
        )
    )
    costing_block.annual_costs.append(
        {"name": opt_name, "annual_cost": stream.annual_cost}
    )



import pyomo.environ as pyo
from ..util import register_costing_parameter_block, make_capital_cost_var


def build_steam_ejector_cost_param_block(blk):
    """
    Build the costing parameter block for the steam ejector.
    """
    blk.base_cost = pyo.Var(
        initialize=1949,
        doc="Base cost coefficient for steam ejector",
        units=pyo.units.USD_2020,
    )
    blk.cost_exponent = pyo.Var(
        initialize=0.3,
        doc="Cost scaling exponent",
        units=pyo.units.dimensionless,
    )

    blk.steam_cost = pyo.Var(
        initialize=0.008,
        units=pyo.units.USD_2018 / pyo.units.kg,
        doc="Steam cost per kg",
    )

    blk.parent_block().register_flow_type("steam", blk.steam_cost)


@register_costing_parameter_block(
    build_rule=build_steam_ejector_cost_param_block,
    parameter_block_name="steam_ejector",
)
def cost_steam_ejector(blk, cost_steam_flow=False):
    """
    Thermo Compressor (Steam Ejector) Costing Method.

    Capital cost is calculated using the equation:
        Capital Cost (USD) = 1949 × (S + EV)^0.3  (Gabriel 2015, Desalination)
    where:
        S  = Motive steam flow rate (kg/h)
        EV = Entrained vapor flow rate (kg/h)

    """

    make_capital_cost_var(blk)
    blk.costing_package.add_cost_factor(blk, "TIC")

    S = pyo.units.convert(
        blk.unit_model.properties_motive_steam[0.0].flow_mass_phase_comp["Vap", "H2O"],
        to_units=pyo.units.kg / pyo.units.hour,
    )

    EV = pyo.units.convert(
        blk.unit_model.properties_entrained_vapor[0.0].flow_mass_phase_comp[
            "Vap", "H2O"
        ],
        to_units=pyo.units.kg / pyo.units.hour,
    )

    blk.capital_cost_constraint = pyo.Constraint(
        expr=blk.capital_cost
        == blk.cost_factor
        * pyo.units.convert(
            blk.costing_package.steam_ejector.base_cost
            * (S + EV) ** blk.costing_package.steam_ejector.cost_exponent,
            to_units=blk.costing_package.base_currency,
        )
    )

    if cost_steam_flow:
        blk.costing_package.cost_flow(
            pyo.units.convert(
                blk.unit_model.properties_motive_steam[0.0].flow_mass_phase_comp[
                    "Vap", "H2O"
                ],
                to_units=pyo.units.kg / pyo.units.s,
            ),
            "steam",
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest
import pyomo.environ as pyo
import idaes.core as idc
import idaes.core.util.model_statistics as istat

from pyomo.util.check_units import assert_units_consistent
from watertap.core.solvers import get_solver
from watertap.costing.watertap_costing_package import WaterTAPCosting
from watertap.costing.util import (
    cost_rectifier,
    register_costing_parameter_block,
)

solver = get_solver()


def build_my_parameter_block(blk):
    blk.v = pyo.Var(initialize=2)


def build_my_parameter_block_3(blk):
    blk.v = pyo.Var(initialize=33)


@register_costing_parameter_block(
    build_rule=build_my_parameter_block,
    parameter_block_name="my_parameter_block",
)
def my_costing_method1(blk):
    pass


@register_costing_parameter_block(
    build_rule=build_my_parameter_block,
    parameter_block_name="my_parameter_block",
)
def my_costing_method2(blk):
    pass


@register_costing_parameter_block(
    build_rule=build_my_parameter_block_3,
    parameter_block_name="my_parameter_block",
)
def my_costing_method3(blk):
    pass


@register_costing_parameter_block(
    build_rule=build_my_parameter_block_3,
    parameter_block_name="parameter_block",
)
def my_costing_method4(blk):
    pass


@pytest.mark.component
def test_register_costing_parameter_block():

    m = pyo.ConcreteModel()
    m.fs = idc.FlowsheetBlock(dynamic=False)

    m.fs.costing = WaterTAPCosting()

    m.fs.unit1 = idc.UnitModelBlock()
    m.fs.unit1.costing = idc.UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=my_costing_method1,
    )

    # test that the parameters are created
    assert isinstance(m.fs.costing.my_parameter_block, pyo.Block)

    # test that the parameters are created
    assert m.fs.costing.my_parameter_block.v.value == 2

    # test that the parameters are fixed
    for v in m.fs.costing.my_parameter_block.component_data_objects(pyo.Var):
        assert v.fixed

    m.fs.costing.my_parameter_block.v.value = 42

    m.fs.unit2 = idc.UnitModelBlock()
    # test that setting the same parameter block does nothing
    m.fs.unit2.costing = idc.UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=my_costing_method2,
    )
    assert m.fs.costing.my_parameter_block.v.value == 42

    # test that resetting the same parameter block raises an error
    m.fs.unit3 = idc.UnitModelBlock()
    with pytest.raises(
        RuntimeError,
        match="Attempting to add identically named costing parameter blocks with "
        "different build rules to the costing package ",
    ):
        m.fs.unit3.costing = idc.UnitModelCostingBlock(
            flowsheet_costing_block=m.fs.costing,
            costing_method=my_costing_method3,
        )

    m.fs.costing.parameter_block = pyo.Block()

    # test that resetting a parameter block added elsewhere raises an error
    with pytest.raises(
        RuntimeError,
        match="Use the register_costing_parameter_block decorator for specifying"
        "costing-package-level parameters",
    ):
        m.fs.unit3.costing = idc.UnitModelCostingBlock(
            flowsheet_costing_block=m.fs.costing,
            costing_method=my_costing_method4,
        )


def build_dummy_cost_rectifier(blk):
    pass


@register_costing_parameter_block(
    build_rule=build_dummy_cost_rectifier,
    parameter_block_name="dummy_rectifier",
)
def dummy_cost_rectifier(blk):
    cost_rectifier(blk)
    blk.costing_package.add_cost_factor(blk, "TIC")
    blk.capital_cost_constraint = pyo.Constraint(
        expr=blk.capital_cost == blk.capital_cost_rectifier
    )


@pytest.mark.component
def test_rectifier_costing():

    # build generic models
    m = pyo.ConcreteModel()
    m.fs = idc.FlowsheetBlock(dynamic=False)
    m.fs.costing = WaterTAPCosting()
    # change base units for value assertions
    m.fs.costing.base_currency = pyo.units.USD_2021
    m.fs.unit = idc.UnitModelBlock()

    # build power variable required for cost_rectifier
    m.fs.unit.power = pyo.Var(
        initialize=100,
        domain=pyo.NonNegativeReals,
        units=pyo.units.kW,
    )
    m.fs.unit.power.fix()

    # cost rectifier
    m.fs.unit.costing = idc.UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=dummy_cost_rectifier,
    )

    # cost process for cost_flow to be applied
    m.fs.costing.cost_process()

    # check model and solve
    assert_units_consistent(m)
    assert istat.degrees_of_freedom(m) == 0
    results = solver.solve(m)
    assert pyo.check_optimal_termination(results)

    # check values
    assert pytest.approx(59320, rel=1e-3) == pyo.value(m.fs.unit.costing.capital_cost)
    assert pytest.approx(111.1, rel=1e-3) == pyo.value(m.fs.unit.costing.ac_power)
    assert pytest.approx(111.1, rel=1e-3) == pyo.value(
        m.fs.costing.aggregate_flow_electricity
    )
    assert pytest.approx(80040, rel=1e-3) == pyo.value(
        m.fs.costing.aggregate_flow_costs["electricity"]
    )



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pytest
from pyomo.environ import (
    ConcreteModel,
)

from idaes.core import (
    FlowsheetBlock,
)

from watertap.core.solvers import get_solver
from idaes.core import UnitModelCostingBlock

from watertap.unit_models.tests.unit_test_harness import UnitTestHarness
import idaes.core.util.scaling as iscale

from watertap.unit_models.crystallizer import Crystallization
import watertap.property_models.unit_specific.cryst_prop_pack as props
from watertap.costing import WaterTAPCosting, CrystallizerCostType

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


# -----------------------------------------------------------------------------
def build():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = props.NaClParameterBlock()

    m.fs.unit = Crystallization(property_package=m.fs.properties)

    # fully specify system
    feed_flow_mass = 1
    feed_mass_frac_NaCl = 0.2126
    feed_mass_frac_H2O = 1 - feed_mass_frac_NaCl
    feed_pressure = 101325
    feed_temperature = 273.15 + 20
    eps = 1e-6
    crystallizer_temperature = 273.15 + 55
    crystallizer_yield = 0.40

    # Fully define feed
    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "NaCl"].fix(
        feed_flow_mass * feed_mass_frac_NaCl
    )
    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
        feed_flow_mass * feed_mass_frac_H2O
    )
    m.fs.unit.inlet.flow_mass_phase_comp[0, "Sol", "NaCl"].fix(eps)
    m.fs.unit.inlet.flow_mass_phase_comp[0, "Vap", "H2O"].fix(eps)
    m.fs.unit.inlet.pressure[0].fix(feed_pressure)
    m.fs.unit.inlet.temperature[0].fix(feed_temperature)

    # Define operating conditions
    m.fs.unit.temperature_operating.fix(crystallizer_temperature)
    m.fs.unit.crystallization_yield["NaCl"].fix(crystallizer_yield)

    # Fix growth rate, crystal length and Sounders brown constant to default values
    m.fs.unit.crystal_growth_rate.fix()
    m.fs.unit.souders_brown_constant.fix()
    m.fs.unit.crystal_median_length.fix()

    iscale.set_scaling_factor(m.fs.unit.pressure_operating, 1e-3)
    iscale.set_scaling_factor(m.fs.unit.properties_out[0].pressure, 1e-5)
    iscale.set_scaling_factor(m.fs.unit.properties_solids[0].pressure, 1e-5)
    iscale.set_scaling_factor(m.fs.unit.properties_vapor[0].pressure, 1e-3)
    iscale.set_scaling_factor(m.fs.unit.properties_out[0].flow_vol_phase["Liq"], 1e3)
    iscale.set_scaling_factor(m.fs.unit.properties_out[0].flow_vol_phase["Vap"], 1e8)
    iscale.set_scaling_factor(m.fs.unit.properties_out[0].pressure_sat, 1e-3)
    iscale.set_scaling_factor(m.fs.unit.properties_solids[0].flow_vol_phase["Vap"], 1e8)
    iscale.set_scaling_factor(m.fs.unit.properties_solids[0].flow_vol_phase["Sol"], 1e8)
    iscale.set_scaling_factor(
        m.fs.unit.properties_vapor[0].dens_mass_solvent["Vap"], 1e1
    )
    iscale.set_scaling_factor(m.fs.unit.properties_out[0].flow_vol_phase["Sol"], 1e12)
    iscale.set_scaling_factor(
        m.fs.unit.properties_solids[0].flow_vol_phase["Liq"], 1e11
    )
    iscale.set_scaling_factor(m.fs.unit.properties_vapor[0].flow_vol_phase["Liq"], 1e12)
    iscale.set_scaling_factor(m.fs.unit.properties_vapor[0].flow_vol_phase["Sol"], 1e12)

    iscale.calculate_scaling_factors(m.fs.unit)

    return m


def build_2():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = props.NaClParameterBlock()

    m.fs.unit = Crystallization(property_package=m.fs.properties)

    # fully specify system
    feed_flow_mass = 1
    feed_mass_frac_NaCl = 0.2126
    feed_mass_frac_H2O = 1 - feed_mass_frac_NaCl
    feed_pressure = 101325
    feed_temperature = 273.15 + 20
    eps = 1e-6
    crystallizer_temperature = 273.15 + 55
    crystallizer_yield = 0.40

    # Fully define feed
    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "NaCl"].fix(
        feed_flow_mass * feed_mass_frac_NaCl
    )
    m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
        feed_flow_mass * feed_mass_frac_H2O
    )
    m.fs.unit.inlet.flow_mass_phase_comp[0, "Sol", "NaCl"].fix(eps)
    m.fs.unit.inlet.flow_mass_phase_comp[0, "Vap", "H2O"].fix(eps)
    m.fs.unit.inlet.pressure[0].fix(feed_pressure)
    m.fs.unit.inlet.temperature[0].fix(feed_temperature)

    # Define operating conditions
    m.fs.unit.temperature_operating.fix(crystallizer_temperature)
    m.fs.unit.crystallization_yield["NaCl"].fix(crystallizer_yield)

    # Fix growth rate, crystal length and Sounders brown constant to default values
    m.fs.unit.crystal_growth_rate.fix()
    m.fs.unit.souders_brown_constant.fix()
    m.fs.unit.crystal_median_length.fix()

    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e-1, index=("Liq", "H2O")
    )
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e-1, index=("Liq", "NaCl")
    )
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e-1, index=("Vap", "H2O")
    )
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e-1, index=("Sol", "NaCl")
    )

    iscale.set_scaling_factor(m.fs.unit.properties_out[0].flow_vol_phase["Sol"], 1e12)
    iscale.set_scaling_factor(m.fs.unit.properties_out[0].flow_vol_phase["Vap"], 1e8)
    iscale.set_scaling_factor(
        m.fs.unit.properties_solids[0].flow_vol_phase["Liq"], 1e11
    )
    iscale.set_scaling_factor(m.fs.unit.properties_solids[0].flow_vol_phase["Vap"], 1e8)
    iscale.set_scaling_factor(m.fs.unit.properties_vapor[0].flow_vol_phase["Liq"], 1e11)
    iscale.set_scaling_factor(m.fs.unit.properties_vapor[0].flow_vol_phase["Sol"], 1e12)

    iscale.calculate_scaling_factors(m.fs.unit)

    return m


@pytest.mark.requires_idaes_solver
class TestCrystallizer(UnitTestHarness):
    def configure(self):
        m = build()

        m.fs.costing = WaterTAPCosting()
        m.fs.unit.costing = UnitModelCostingBlock(
            flowsheet_costing_block=m.fs.costing,
            costing_method_arguments={"cost_type": CrystallizerCostType.mass_basis},
        )
        m.fs.costing.cost_process()
        iscale.set_scaling_factor(m.fs.unit.costing.capital_cost, 1e-5)

        self.unit_solutions[m.fs.unit.solids.flow_mass_phase_comp[0, "Sol", "NaCl"]] = (
            0.08504096
        )
        self.unit_solutions[m.fs.unit.outlet.flow_mass_phase_comp[0, "Liq", "NaCl"]] = (
            0.12756
        )
        self.unit_solutions[m.fs.unit.outlet.flow_mass_phase_comp[0, "Liq", "H2O"]] = (
            0.34570767
        )
        self.unit_solutions[
            m.fs.unit.properties_out[0].mass_frac_phase_comp["Liq", "NaCl"]
        ] = 0.26953035
        self.unit_solutions[m.fs.unit.pressure_operating] = 11991.82426
        self.unit_solutions[m.fs.unit.work_mechanical[0]] = 1127.222711
        self.unit_solutions[m.fs.unit.diameter_crystallizer] = 1.204904
        self.unit_solutions[m.fs.unit.volume_suspension] = 1.6194469
        self.unit_solutions[m.fs.unit.t_res] = 1.022821
        self.unit_solutions[m.fs.unit.costing.capital_cost] = 600166.827

        comp_lst = ["NaCl", "H2O"]
        phase_lst = ["Sol", "Liq", "Vap"]
        phase_comp_list = [
            (p, j)
            for j in comp_lst
            for p in phase_lst
            if (p, j) in m.fs.unit.properties_in[0].phase_component_set
        ]

        flow_mass_in = sum(
            m.fs.unit.properties_in[0].flow_mass_phase_comp[p, j]
            for p in phase_lst
            for j in comp_lst
            if (p, j) in phase_comp_list
        )
        flow_mass_out = sum(
            m.fs.unit.properties_out[0].flow_mass_phase_comp[p, j]
            for p in phase_lst
            for j in comp_lst
            if (p, j) in phase_comp_list
        )
        flow_mass_solids = sum(
            m.fs.unit.properties_solids[0].flow_mass_phase_comp[p, j]
            for p in phase_lst
            for j in comp_lst
            if (p, j) in phase_comp_list
        )
        flow_mass_vapor = sum(
            m.fs.unit.properties_vapor[0].flow_mass_phase_comp[p, j]
            for p in phase_lst
            for j in comp_lst
            if (p, j) in phase_comp_list
        )

        kj_to_j = 1e3
        self.conservation_equality = {
            "Check 1": {
                "in": flow_mass_in,
                "out": flow_mass_out + flow_mass_solids + flow_mass_vapor,
            },
            "Check 2": {
                "in": flow_mass_in,
                "out": (
                    flow_mass_out * m.fs.unit.properties_out[0].enth_mass_phase["Liq"]
                    + flow_mass_vapor
                    * m.fs.unit.properties_vapor[0].enth_mass_solvent["Vap"]
                    + flow_mass_solids
                    * m.fs.unit.properties_solids[0].enth_mass_solute["Sol"]
                    + flow_mass_solids
                    * m.fs.unit.properties_solids[0].dh_crystallization_mass_comp[
                        "NaCl"
                    ]
                    - m.fs.unit.work_mechanical[0] * kj_to_j
                )
                / m.fs.unit.properties_in[0].enth_mass_phase["Liq"],
            },
        }

        return m


@pytest.mark.requires_idaes_solver
class TestCrystallizer_costing_by_mass(UnitTestHarness):
    def configure(self):
        m = build()
        m.fs.unit.crystal_growth_rate.fix(5e-8)
        m.fs.unit.souders_brown_constant.fix(0.0244)
        m.fs.unit.crystal_median_length.fix(0.4e-3)

        m.fs.costing = WaterTAPCosting()
        m.fs.unit.costing = UnitModelCostingBlock(
            flowsheet_costing_block=m.fs.costing,
            costing_method_arguments={"cost_type": CrystallizerCostType.mass_basis},
        )
        m.fs.costing.cost_process()
        iscale.set_scaling_factor(m.fs.unit.costing.capital_cost, 1e-5)

        self.unit_solutions[m.fs.unit.diameter_crystallizer] = 1.5427211
        self.unit_solutions[m.fs.unit.volume_suspension] = 0.95871257
        self.unit_solutions[m.fs.unit.costing.capital_cost] = 600166.827

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "NaCl"]
                + m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "H2O"]
                + m.fs.unit.inlet.flow_mass_phase_comp[0, "Sol", "NaCl"]
                + m.fs.unit.inlet.flow_mass_phase_comp[0, "Vap", "H2O"],
                "out": m.fs.unit.outlet.flow_mass_phase_comp[0, "Liq", "NaCl"]
                + m.fs.unit.outlet.flow_mass_phase_comp[0, "Liq", "H2O"]
                + m.fs.unit.solids.flow_mass_phase_comp[0, "Sol", "NaCl"]
                + m.fs.unit.vapor.flow_mass_phase_comp[0, "Vap", "H2O"],
            },
        }

        return m


@pytest.mark.requires_idaes_solver
class TestCrystallizer_costing_by_volume(UnitTestHarness):
    def configure(self):
        m = build_2()
        m.fs.unit.crystal_growth_rate.fix(5e-8)
        m.fs.unit.souders_brown_constant.fix(0.0244)
        m.fs.unit.crystal_median_length.fix(0.4e-3)

        m.fs.costing = WaterTAPCosting()
        m.fs.unit.costing = UnitModelCostingBlock(
            flowsheet_costing_block=m.fs.costing,
            costing_method_arguments={"cost_type": CrystallizerCostType.volume_basis},
        )

        m.fs.costing.crystallizer.steam_pressure.fix(5)
        m.fs.costing.crystallizer.NaCl_recovery_value.fix(-0.07)

        m.fs.costing.cost_process()

        iscale.set_scaling_factor(m.fs.unit.costing.capital_cost, 1e-5)

        self.unit_solutions[m.fs.unit.diameter_crystallizer] = 1.5427211
        self.unit_solutions[m.fs.unit.volume_suspension] = 0.9587126
        self.unit_solutions[m.fs.unit.costing.capital_cost] = 398225.15644

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "NaCl"]
                + m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "H2O"]
                + m.fs.unit.inlet.flow_mass_phase_comp[0, "Sol", "NaCl"]
                + m.fs.unit.inlet.flow_mass_phase_comp[0, "Vap", "H2O"],
                "out": m.fs.unit.outlet.flow_mass_phase_comp[0, "Liq", "NaCl"]
                + m.fs.unit.outlet.flow_mass_phase_comp[0, "Liq", "H2O"]
                + m.fs.unit.solids.flow_mass_phase_comp[0, "Sol", "NaCl"]
                + m.fs.unit.vapor.flow_mass_phase_comp[0, "Vap", "H2O"],
            },
        }

        return m



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest

from pyomo.environ import assert_optimal_termination, value
from pyomo.util.check_units import assert_units_consistent
from idaes.core import MaterialFlowBasis
from watertap.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from watertap.property_models.multicomp_aq_sol_prop_pack import DiffusivityCalculation
from watertap.unit_models.gac import (
    FilmTransferCoefficientType,
    SurfaceDiffusionCoefficientType,
)
from watertap.costing.unit_models.gac import ContactorType
from watertap.flowsheets.gac import gac as gac_fs

__author__ = "Hunter Barber"

solver = get_solver()


class TestGACFlowsheet:
    @pytest.fixture(scope="class")
    def gac_frame(self):

        m = gac_fs.build(
            material_flow_basis="molar",
            film_transfer_coefficient_type="calculated",
            surface_diffusion_coefficient_type="calculated",
            diffusivity_calculation="HaydukLaudie",
            cost_contactor_type="pressure",
        )

        return m

    @pytest.mark.requires_idaes_solver
    @pytest.mark.component
    def test_solve(self, gac_frame):

        m = gac_frame
        gac_fs.initialize(m)
        res = gac_fs.optimize(m)

        assert_units_consistent(m)
        assert degrees_of_freedom(m) == 0
        assert_optimal_termination(res)

        feed = m.fs.gac.process_flow.properties_in[0]
        cost = m.fs.costing
        assert value(feed.flow_vol_phase["Liq"]) == pytest.approx(0.043813, rel=1e-3)
        assert value(feed.flow_mass_phase_comp["Liq", "solute"]) == pytest.approx(
            0.0043813, rel=1e-3
        )
        assert value(feed.flow_mol_phase_comp["Liq", "solute"]) == pytest.approx(
            0.05477, rel=1e-3
        )
        assert value(m.fs.gac.operational_time) == pytest.approx(3207000, rel=1e-3)
        assert value(cost.total_capital_cost) == pytest.approx(744000, rel=1e-3)
        assert value(cost.total_operating_cost) == pytest.approx(626900, rel=1e-3)
        assert value(cost.LCOW) == pytest.approx(0.5636, rel=1e-3)

    @pytest.mark.requires_idaes_solver
    @pytest.mark.component
    def test_build_solve_options(self):

        # test build and solve at initial conditions for all config options
        for flow_basis in [MaterialFlowBasis.molar, MaterialFlowBasis.mass]:
            for film_option in FilmTransferCoefficientType:
                for surface_option in SurfaceDiffusionCoefficientType:
                    for diffus_option in DiffusivityCalculation:
                        for cost_option in ContactorType:

                            m = gac_fs.build(
                                material_flow_basis=flow_basis.name,
                                film_transfer_coefficient_type=film_option.name,
                                surface_diffusion_coefficient_type=surface_option.name,
                                diffusivity_calculation=diffus_option.name,
                                cost_contactor_type=cost_option.name,
                            )
                            gac_fs.initialize(m)
                            res = gac_fs.optimize(m)

                            assert_units_consistent(m)
                            assert degrees_of_freedom(m) == 0
                            assert_optimal_termination(res)



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import os
import pytest
from pyomo.environ import (
    ConcreteModel,
    Var,
    Constraint,
    units as pyunits,
    assert_optimal_termination,
)
from pyomo.util.check_units import assert_units_consistent
from idaes.core import FlowsheetBlock
from idaes.core.util.exceptions import ConfigurationError
from watertap.property_models.water_prop_pack import WaterParameterBlock
from watertap.property_models.multicomp_aq_sol_prop_pack import (
    MCASParameterBlock,
    MaterialFlowBasis,
)
from idaes.core.util.model_statistics import degrees_of_freedom

from idaes.core.surrogate.surrogate_block import SurrogateBlock
from idaes.core.surrogate.pysmo_surrogate import (
    PysmoSurrogate,
)
from watertap.unit_models.surrogate_crystallizer import SurrogateCrystallizer

from idaes.core import UnitModelCostingBlock
from watertap.costing import WaterTAPCosting
from watertap.core.solvers import get_solver

__author__ = "Oluwamayowa Amusat, Adam Atia"


# TODO: can consider adding NN model surrogate function and test
def add_crystallizer_rbf_model(
    blk,
    surrogate_inputs_with_bounds,
    surrogate_outputs,
    surrogate_to_flowsheet_basis_ratio,
):
    ##############################################################################################################
    # block loading into IDAES Surrogate Block
    ##############################################################################################################
    crystallizer_inputs = [
        surrogate_inputs_with_bounds[k]["flowsheet_var"]
        for k in surrogate_inputs_with_bounds.keys()
    ]
    crystallizer_outputs = [
        surrogate_outputs[k]["flowsheet_var"] for k in surrogate_outputs.keys()
    ]

    try:
        assert set(list(blk.solids_list)).issubset(surrogate_outputs.keys())
    except AssertionError:
        raise ConfigurationError(
            "List of solids provided in crystallizer model must match surrogate output keys for solids. Please check."
        )

    filename = [
        "Vapor_Pressure",
        "Calcite_g",
        "Anhydrite_g",
        "Glauberite_g",
        "Halite_g",
    ]

    for sm in range(0, len(filename)):
        block_name = "crystallizer_surrogate" + "_" + filename[sm]
        blk.add_component(block_name, SurrogateBlock(concrete=True))
        surrogate_name = f"{filename[sm]}.json"
        surrogate_path = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "..",
            "..",
            "data",
            "surrogate_defaults",
            "surrogate_crystallizer_defaults",
            surrogate_name,
        )
        current_surrogate = PysmoSurrogate.load_from_file(surrogate_path)

        getattr(blk, block_name).build_model(
            current_surrogate,
            input_vars=crystallizer_inputs,
            output_vars=crystallizer_outputs[sm],
        )

    # Add constraints tying provided mass variables with crystallizer unit variables - done to avoid unit challenges
    def eq_mass_conversion_constraint(b, j):
        return (
            b.flow_mass_sol_comp_apparent[j]
            == pyunits.convert(
                surrogate_outputs[j]["flowsheet_var"], to_units=pyunits.kg / pyunits.s
            )
            / surrogate_to_flowsheet_basis_ratio
        )

    blk.mass_conversion_constraints = Constraint(
        blk.solids_list, rule=eq_mass_conversion_constraint
    )

    # Add a constraint to tie pressures together after unit conversion
    blk.pressure_conversion_constraints = Constraint(
        expr=blk.pressure_operating
        == pyunits.convert(
            surrogate_outputs["Vapor Pressure (atm)"]["flowsheet_var"],
            to_units=pyunits.Pa,
        )
    )


@pytest.mark.component
def test_rbf_surrogate():
    m = ConcreteModel()
    m.case = "BGW1"
    m.fs = FlowsheetBlock(dynamic=False)

    input_ions = ["Cl_-", "Na_+", "SO4_2-", "Mg_2+", "Ca_2+", "K_+", "HCO3_-"]
    solids_list = {
        "Calcite_g": {"Ca_2+": 1, "HCO3_-": 1},
        "Anhydrite_g": {"Ca_2+": 1, "SO4_2-": 1},
        "Glauberite_g": {"Ca_2+": 1, "Na_+": 2, "SO4_2-": 2},
        "Halite_g": {"Na_+": 1, "Cl_-": 1},
    }
    m.fs.cryst_prop_feed = MCASParameterBlock(
        solute_list=input_ions,
        material_flow_basis=MaterialFlowBasis.mass,
    )

    m.fs.water_properties_vapor = WaterParameterBlock()

    # Create crystallizer framework
    m.fs.cryst = SurrogateCrystallizer(
        property_package=m.fs.cryst_prop_feed,
        vapor_property_package=m.fs.water_properties_vapor,
        solids_ions_dict=solids_list,
    )

    # Feed composition
    feed = {
        "ion_composition_g_kg": {
            "Na_+": 91.49859876,
            "K_+": 3.185931759,
            "Cl_-": 159.0941278,
            "Ca_2+": 0.788476605,
            "Mg_2+": 10.5806906,
            "HCO3_-": 1.614773395,
            "SO4_2-": 22.20893266,
        },
        "water_content_kg": 12.0764972138554,
    }

    g_to_kg = 1e-3
    for i in m.fs.cryst_prop_feed.component_list:
        if i == "H2O":
            m.fs.cryst.inlet.flow_mass_phase_comp[0.0, "Liq", i].fix(
                feed["water_content_kg"]
            )
        else:
            m.fs.cryst.inlet.flow_mass_phase_comp[0.0, "Liq", i].fix(
                feed["water_content_kg"] * feed["ion_composition_g_kg"][i] * g_to_kg
            )

    m.fs.cryst.inlet.temperature.fix(298.15)
    m.fs.cryst.inlet.pressure.fix(101325)

    assert_units_consistent(m)

    # Create dummy variables for inputs and outputs
    surrogate_inputs = {
        "Temperature": {
            "flowsheet_var": m.fs.cryst.temperature_operating,
            "var_bounds": (303.15, 372.15),
        },
        "Evaporation percent": {
            "flowsheet_var": m.fs.cryst.evaporation_percent,
            "var_bounds": (30.0, 98.0),
        },
    }
    m.fs.o1 = Var(units=pyunits.atm)
    m.fs.o2 = Var(units=pyunits.g / pyunits.s)
    m.fs.o3 = Var(units=pyunits.g / pyunits.s)
    m.fs.o4 = Var(units=pyunits.g / pyunits.s)
    m.fs.o5 = Var(units=pyunits.g / pyunits.s)
    surrogate_outputs = {
        "Vapor Pressure (atm)": {"flowsheet_var": m.fs.o1, "Solid": False},
        "Calcite_g": {"flowsheet_var": m.fs.o2, "Solid": True},
        "Anhydrite_g": {"flowsheet_var": m.fs.o3, "Solid": True},
        "Glauberite_g": {"flowsheet_var": m.fs.o4, "Solid": True},
        "Halite_g": {"flowsheet_var": m.fs.o5, "Solid": True},
    }

    add_crystallizer_rbf_model(
        m.fs.cryst,
        surrogate_inputs_with_bounds=surrogate_inputs,
        surrogate_outputs=surrogate_outputs,
        surrogate_to_flowsheet_basis_ratio=1,
    )

    # Costing
    m.fs.costing = WaterTAPCosting()
    m.fs.cryst.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
    )
    m.fs.costing.cost_process()

    # Edit bound on liquid pressures from property package
    m.fs.cryst.properties_out_liq[0].pressure.setlb(1000)

    # 1. Simulate single case
    m.fs.cryst.temperature_operating.fix(40 + 273.15)
    m.fs.cryst.evaporation_percent.fix(60)

    # Initialize and solve
    assert degrees_of_freedom(m) == 0
    m.fs.cryst.initialize()
    solver = get_solver()
    res = solver.solve(m)
    assert_optimal_termination(res)
    m.fs.cryst.report()



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

###############################################################################
#
# OLI Systems, Inc. Copyright © 2022, all rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation and/or
# other materials provided with the distribution.
#
# 3. Neither the name of OLI Systems, Inc. nor the names of any contributors to
# the software made available herein may be used to endorse or promote products derived
# from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
# SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
# OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# You are under no obligation whatsoever to provide any bug fixes, patches, or upgrades to the
# features, functionality or performance of the source code ("Enhancements") to anyone; however,
# if you choose to make your Enhancements available either publicly, or directly to OLI Systems, Inc.,
# without imposing a separate written license agreement for such Enhancements, then you hereby grant
# the following license: a non-exclusive, royalty-free perpetual license to install, use, modify, prepare
# derivative works, incorporate into other computer software, distribute, and sublicense such enhancements
# or derivative works thereof, in binary and source code form.
###############################################################################
__author__ = "Paul Vecchiarelli"

import logging

from copy import deepcopy

import json
from pathlib import Path
from datetime import datetime, timedelta, timezone

from pyomo.common.dependencies import attempt_import

cryptography, cryptography_available = attempt_import(
    "cryptography", defer_import=False
)
if cryptography_available:
    from cryptography.fernet import Fernet
requests, requests_available = attempt_import("requests", defer_import=False)

_logger = logging.getLogger(__name__)
# set to info level, so user can see what is going on
handler = logging.StreamHandler()
formatter = logging.Formatter(
    "OLIAPI - %(asctime)s - %(levelname)s - %(message)s", "%H:%M:%S"
)
handler.setFormatter(formatter)
_logger.addHandler(handler)
_logger.setLevel(logging.DEBUG)


class CredentialManager:
    """
    A class to handle credentials for OLI Cloud.
    """

    def __init__(
        self,
        username="",
        password="",
        root_url="",
        auth_url="",
        config_file="./credentials.txt",
        encryption_key="",
        access_keys=[],
        test=False,
        interactive_mode=True,
        refresh=False,
        debug_level="INFO",
    ):
        """
        Manages credentials for OLIApi authentication requests.

        :param username: user's username
        :param password: user's password
        :param root_url: root url
        :param auth_url: authorization url
        :param config_file: existing/desired path (absolute, or relative to the working directory) to encrypted oli_config_file
        :param encryption_key: fernet key generated by credential manager object
        :param access_keys: list of access keys generated by user
        :param test: bool switch for automation during tests
        :param interactive_mode: enables direct interaction with user through prompts
        :param refresh: bool to use refresh token in login method
        :param debug_level: string defining level of logging activity

        """

        self.test = test
        self.access_key = ""
        self.encryption_key = encryption_key
        self.config_file = Path(config_file).resolve()
        self.refresh = refresh
        self.interactive_mode = interactive_mode
        self.debug_level = debug_level
        if self.debug_level == "INFO":
            _logger.setLevel(logging.INFO)
        else:
            _logger.setLevel(logging.DEBUG)
        self._manage_credentials(
            username,
            password,
            root_url,
            auth_url,
            access_keys,
        )
        self.set_headers()

    def set_headers(self):
        """
        Creates OLI Cloud API headers and performs initial login.
        """

        if self.access_key:
            self.headers = {"authorization": "API-KEY " + self.access_key}
            self.login()
        else:
            self.login()
            self.headers = {"authorization": "Bearer " + self.jwt_token}

    def update_headers(self, new_header):
        """
        Updates existing headers with new header.

        :param new_header: dict containing new header key and value

        :return updated_headers: dict containing updated headers
        """

        updated_headers = deepcopy(self.headers)
        updated_headers.update(new_header)
        return updated_headers

    def _manage_credentials(self, username, password, root_url, auth_url, access_keys):
        """
        Method to save/load OLI credentials.

        :param username: user's username
        :param password: user's password
        :param root_url: root url
        :param auth_url: authorization url
        :param access_keys: list of access keys generated by user
        """

        if not cryptography_available:
            raise ModuleNotFoundError("Module 'cryptography' not available.")

        if self.encryption_key:
            if self.config_file.is_file():
                self.credentials = self._decrypt_credentials()
            else:
                raise OSError(
                    f" Config file {self.config_file} does not exist."
                    + " Provide login credentials to generate encrypted file."
                )
        else:
            self.credentials = {
                "username": username,
                "password": password,
                "root_url": root_url,
                "auth_url": auth_url,
                "access_keys": access_keys,
            }
            if not access_keys:
                self._check_credentials(
                    ["username", "password", "root_url", "auth_url"]
                )
            else:
                self._check_credentials(["root_url", "access_keys"])

            if self._write_permission():
                self.encryption_key = self._encrypt_credentials()

        self.access_key = self.set_access_key()

        self.dbs_url = self.credentials["root_url"] + "/channel/dbs"
        self.upload_dbs_url = self.credentials["root_url"] + "/channel/upload/dbs"
        self.access_key_url = self.credentials["root_url"] + "/user/api-key"
        self.engine_url = self.credentials["root_url"] + "/engine"
        self._delete_dbs_url = self.credentials["root_url"] + "/channel/file/"

    def _decrypt_credentials(self):
        """
        Basic decryption method for credentials.

        :return credentials: login credentials for OLI Cloud
        """

        try:
            with open(self.config_file, "rb") as f:
                encrypted_credentials = f.read()

            cipher = Fernet(self.encryption_key)
            decrypted_credentials = cipher.decrypt(encrypted_credentials).decode()
            credentials = json.loads(decrypted_credentials)
            return credentials

        except:
            raise RuntimeError(" Failed decryption.")

    def _check_credentials(self, keys):
        """
        Check to see if required credentials are missing.

        :param keys: keys required for login method
        """
        e = [k for k, v in self.credentials.items() if k in keys if not v]
        if e:
            raise IOError(f" Incomplete credentials for the following keys: {e}.")

    def _write_permission(self):
        """
        Ensures user permits deletion of specified files.

        :return bool: status of user permission (to write encrypted config_file to disk)
        """

        if self.test:
            return True
        else:
            if self.interactive_mode:
                r = input(
                    "WaterTAP will write encrypted file to store OLI Cloud credentials: [y]/n: "
                )
                if (r.lower() == "y") or (r == ""):
                    return True
                return False

    # TODO: consider updating credentials when writing, rather than resetting blank ones
    def _encrypt_credentials(self):
        """
        Basic encryption method for credentials.
        """

        encryption_key = Fernet.generate_key()
        _logger.info(f" Secret key is {encryption_key.decode()}")

        try:
            cipher = Fernet(encryption_key)
            encrypted_credentials = cipher.encrypt(
                json.dumps(self.credentials).encode()
            )
            with open(self.config_file, "wb") as f:
                f.write(encrypted_credentials)
            return encryption_key

        except:
            raise RuntimeError(f" Failed encryption.")

    def set_access_key(self):
        """
        Allows access key to be selected from list if more than one is provided.
        """
        if len(self.credentials["access_keys"]) == 0:
            return ""
        elif len(self.credentials["access_keys"]) == 1:
            return self.credentials["access_keys"][0]
        else:
            _logger.info("Specify an access key: ")
            for i in range(len(self.credentials["access_keys"])):
                _logger.info(f"{i}\t{self.credentials['access_keys'][i]}")
                if self.test:
                    r = 0
                else:
                    r = int(input(" "))
            return self.credentials["access_keys"][r]

    # TODO: possibly save api keys as objects with expiry and other attributes
    def generate_oliapi_access_key(self, key_lifetime=365):
        """
        Generate an access key for OLI Cloud.

        :param key_lifetime: integer number of days key will be valid

        :return string: Response text containing the access key information or an error message
        """

        def _set_expiry_timestamp(key_lifetime):
            """
            Set expiry date for OLI Cloud access key.

            :param key_lifetime: integer number of days key will be valid (up to 365)

            :return unix_timestamp_ms: unix timestamp (in ms) for when key will expire
            """

            if key_lifetime < 1 or key_lifetime > 365:
                key_lifetime = 365

            _logger.debug(f"Maximum key lifetime is 365 days, {key_lifetime} provided.")
            current_time = datetime.now(timezone.utc)
            expiry_timestamp = (current_time + timedelta(days=key_lifetime)).timestamp()
            unix_timestamp_ms = int(expiry_timestamp * 1000)
            return unix_timestamp_ms

        response = requests.post(
            self.access_key_url,
            headers=self.update_headers({"Content-Type": "application/json"}),
            data=json.dumps({"expiry": _set_expiry_timestamp(key_lifetime)}),
        )

        self.credentials["access_keys"].append(
            json.loads(response.text)["data"]["apiKey"]
        )
        _logger.info(response.text)
        return response.text

    def delete_oliapi_access_key(self, api_key):
        """
        Delete an access key for OLI Cloud.

        :param api_key: The access key to delete

        :return string: Response text containing the success message or an error message
        """

        response = requests.delete(
            self.access_key_url,
            headers=self.update_headers({"Content-Type": "application/json"}),
            data=json.dumps({"apiKey": api_key}),
        )
        _logger.info(response.text)
        return response.text

    def login(self):
        """
        Log in to OLI Cloud using access key or credentials.

        :return status: bool indicating success or failure
        """

        req_result = ""
        if self.access_key:
            _logger.info("Logging into OLI API using access key")
            req_result = requests.get(
                self.dbs_url,
                headers=self.update_headers(
                    {"Content-Type": "application/x-www-form-urlencoded"}
                ),
            )
        else:
            _logger.info("Logging into OLI API using username and password")
            if self.refresh:
                if hasattr(self, "refresh_token"):
                    body = {
                        "refresh_token": self.refresh_token,
                        "grant_type": "refresh_token",
                        "client_id": "apiclient",
                    }
                else:
                    body = {
                        "username": self.credentials["username"],
                        "password": self.credentials["password"],
                        "grant_type": "password",
                        "client_id": "apiclient",
                    }
                    status = self.auth_status(body, req_result)
                    return status
            else:
                body = {
                    "username": self.credentials["username"],
                    "password": self.credentials["password"],
                    "grant_type": "password",
                    "client_id": "apiclient",
                }
            status = self.auth_status(body, req_result)
            return status

    def auth_status(self, body, req_result=None):
        """
        Posts authorization request to OLI Cloud.

        :param body: dictionary containing authorization data

        :return bool: bool indicating success or failure
        """

        if not req_result:
            req_result = requests.post(
                self.credentials["auth_url"],
                headers={"Content-Type": "application/x-www-form-urlencoded"},
                data=body,
            )
        if req_result.status_code == 200:
            _logger.debug(f"Status code is {req_result.status_code}")
            _logger.info("Log in successful")
            if self.access_key:
                return True
            else:
                req_result = req_result.json()
                if "access_token" in req_result:
                    _logger.debug(f"Login access token: {req_result['access_token']}")
                    self.jwt_token = req_result["access_token"]
                    if "refresh_token" in req_result:
                        _logger.debug(
                            f"Login refresh token: {req_result['refresh_token']}"
                        )
                        self.refresh_token = req_result["refresh_token"]
                        return True
        else:
            raise ConnectionError(
                f" OLI login failed. Status code is {req_result.status_code}.\n"
            )



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest
from pyomo.environ import value
from watertap.core.solvers import get_solver
from watertap.core.util.initialization import check_dof
import watertap.flowsheets.electrodialysis.electrodialysis_1stack_conc_recirc as edfs

__author__ = "Xiangyu Bi"


class TestElectrodialysis1StackFS:
    @pytest.fixture(scope="class")
    def ED1D1Stack_conc_recirc(self):
        m = edfs.build()
        return m

    @pytest.mark.component
    def test_specific_operating_conditions(self, ED1D1Stack_conc_recirc):
        m = ED1D1Stack_conc_recirc
        solver = get_solver()
        # Testing a feeding salinity of 2g/L.
        init_arg = {
            ("flow_vol_phase", ("Liq")): 5.2e-4,
            ("conc_mol_phase_comp", ("Liq", "Na_+")): 34.188,
            ("conc_mol_phase_comp", ("Liq", "Cl_-")): 34.188,
        }
        m.fs.feed.properties.calculate_state(
            init_arg,
            hold_state=True,
        )
        m.fs.EDstack.voltage_applied[0].fix(10)
        m.fs.recovery_vol_H2O.fix(0.7)
        edfs._condition_base(m)
        check_dof(m)

        # Initialize and solve the model
        edfs.initialize_system(m, solver=solver)
        edfs.solve(m, solver=solver)

        assert value(m.fs.feed.properties[0].flow_vol_phase["Liq"]) == pytest.approx(
            5.2e-4, rel=1e-3
        )
        assert value(m.fs.prod.properties[0].flow_vol_phase["Liq"]) == pytest.approx(
            3.64e-4, rel=1e-3
        )
        assert value(m.fs.disp.properties[0].flow_vol_phase["Liq"]) == pytest.approx(
            1.56e-4, rel=1e-3
        )
        assert value(
            sum(
                m.fs.feed.properties[0].conc_mass_phase_comp["Liq", j]
                for j in m.fs.properties.ion_set
            )
        ) == pytest.approx(2.00, rel=1e-3)
        assert value(m.fs.product_salinity) == pytest.approx(1.1333, rel=1e-3)
        assert value(m.fs.disposal_salinity) == pytest.approx(4.0223, rel=1e-3)
        assert value(m.fs.mem_area) == pytest.approx(18.5338, rel=1e-3)
        assert value(m.fs.costing.specific_energy_consumption) == pytest.approx(
            0.1348, abs=0.001
        )
        assert value(m.fs.costing.LCOW) == pytest.approx(0.3996, rel=1e-3)
        assert value(m.fs.EDstack.inlet_concentrate.pressure[0]) == pytest.approx(
            169278.127, rel=1e-3
        )
        assert value(m.fs.EDstack.outlet_concentrate.pressure[0]) == pytest.approx(
            101325.00, rel=1e-3
        )
        assert value(m.fs.EDstack.inlet_diluate.pressure[0]) == pytest.approx(
            169278.127, rel=1e-3
        )
        assert value(m.fs.EDstack.outlet_diluate.pressure[0]) == pytest.approx(
            101325.00, rel=1e-3
        )
        assert value(m.fs.EDstack.inlet_concentrate.temperature[0]) == pytest.approx(
            298.15, rel=1e-3
        )
        assert value(m.fs.EDstack.outlet_concentrate.temperature[0]) == pytest.approx(
            298.15, rel=1e-3
        )
        assert value(m.fs.EDstack.inlet_diluate.temperature[0]) == pytest.approx(
            298.15, rel=1e-3
        )
        assert value(m.fs.EDstack.outlet_diluate.temperature[0]) == pytest.approx(
            298.15, rel=1e-3
        )

    @pytest.mark.component
    def test_optimization(self):
        m = edfs.main()
        assert value(m.fs.feed.properties[0].flow_vol_phase["Liq"]) == pytest.approx(
            5.2e-4, rel=1e-3
        )
        assert value(m.fs.prod.properties[0].flow_vol_phase["Liq"]) == pytest.approx(
            3.64e-4, rel=1e-3
        )
        assert value(m.fs.disp.properties[0].flow_vol_phase["Liq"]) == pytest.approx(
            1.56e-4, rel=1e-3
        )
        assert value(
            sum(
                m.fs.feed.properties[0].conc_mass_phase_comp["Liq", j]
                for j in m.fs.properties.ion_set
            )
        ) == pytest.approx(2.00, rel=1e-3)
        assert value(m.fs.product_salinity) == pytest.approx(0.1000, rel=1e-3)
        assert value(m.fs.disposal_salinity) == pytest.approx(6.4333, rel=1e-3)
        assert value(m.fs.mem_area) == pytest.approx(15.1283, rel=1e-3)
        assert value(m.fs.EDstack.cell_pair_num) == pytest.approx(16, rel=1e-8)
        assert value(m.fs.EDstack.cell_length) == pytest.approx(4.800, rel=1e-3)
        assert value(m.fs.EDstack.voltage_applied[0]) == pytest.approx(18.785, rel=1e-3)
        assert value(m.fs.costing.specific_energy_consumption) == pytest.approx(
            1.7775, rel=1e-3
        )
        assert value(m.fs.costing.LCOW) == pytest.approx(0.6254, rel=1e-3)
        assert value(m.fs.EDstack.inlet_concentrate.pressure[0]) == pytest.approx(
            784785.332, rel=1e-3
        )
        assert value(m.fs.EDstack.outlet_concentrate.pressure[0]) == pytest.approx(
            101325.00, rel=1e-3
        )
        assert value(m.fs.EDstack.inlet_diluate.pressure[0]) == pytest.approx(
            784785.332, rel=1e-3
        )
        assert value(m.fs.EDstack.outlet_diluate.pressure[0]) == pytest.approx(
            101325.00, rel=1e-3
        )
        assert value(m.fs.EDstack.inlet_concentrate.temperature[0]) == pytest.approx(
            298.15, rel=1e-3
        )
        assert value(m.fs.EDstack.outlet_concentrate.temperature[0]) == pytest.approx(
            298.15, rel=1e-3
        )
        assert value(m.fs.EDstack.inlet_diluate.temperature[0]) == pytest.approx(
            298.15, rel=1e-3
        )
        assert value(m.fs.EDstack.outlet_diluate.temperature[0]) == pytest.approx(
            298.15, rel=1e-3
        )
